# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.
"""
String table manager. Internally used in eudplib.
"""

from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Literal

from ... import utils as ut
from ...localize import _
from .chktok import CHK

if TYPE_CHECKING:
    from .stringmap import StringIdMap

unit_name_encoding: str | None = None


def DecodeUnitNameAs(e: str) -> None:  # noqa: N802
    global unit_name_encoding
    "".encode(e)  # check if e is valid encoding
    unit_name_encoding = e


def _ignore_color(s: bytes) -> bytes | None:
    has_color = False
    ret = []
    for x in s:
        if 0x01 <= x <= 0x1F or x == 0x7F:
            has_color = True
        else:
            ret.append(x)
    if has_color:
        return bytes(ret)
    else:
        return None


def _b2in(n: int) -> Callable[[Sequence[int], int], int]:
    b2in_map = {2: ut.b2i2, 4: ut.b2i4}
    return b2in_map[n]


def _i2bn(n: int) -> Callable[[int], bytes]:
    i2bn_map = {2: ut.i2b2, 4: ut.i2b4}
    return i2bn_map[n]


def _roundup_by_4(num: int) -> int:
    return -((-num) // 4) * 4


class TBL:
    def __init__(
        self,
        content: bytes | None = None,
        init_chkt: tuple[CHK, StringIdMap, StringIdMap, StringIdMap] | None = None,
        *,
        load_entry: Literal[2, 4] = 2,
        save_entry: Literal[2, 4] = 4,
    ) -> None:
        self._loadentry: Literal[2, 4] = load_entry
        self._saveentry: Literal[2, 4] = save_entry

        # datatb : table of strings                       : string data table
        # dataindextb : string id -> data id              : string offset table
        # stringmap : string -> representative string id
        self._datatb: list[bytes] = []
        self._stringmap: dict[bytes, int] = {}
        self._dataindextb: list[int] = []  # String starts from #1
        self._capacity: int = save_entry  # Size of STR section
        self._emptystring: list[tuple[int, bytes] | int] = []
        self._loaded: bool = False
        self._first_extended_string: bytes | None = None

        self._finalized: bool = False
        self._tbldata: bytes = b""
        self._stroffset: list[int] = []

        if content is not None:
            if init_chkt:
                self.load_tbl_with_chk(content, init_chkt)
            else:
                self.load_tbl(content)

    def load_tbl(self, content: bytes) -> None:
        # ut.ep_assert(not self._loaded, "String data are already loaded")
        self._datatb.clear()
        self._stringmap.clear()
        self._dataindextb.clear()
        self._emptystring.clear()
        self._first_extended_string = None
        self._capacity = self._saveentry

        size = self._loadentry
        b2i = _b2in(size)
        stringcount = b2i(content, 0)

        for i in range(stringcount):
            i += 1
            stringoffset = b2i(content, i * size)
            send = stringoffset
            while content[send] != 0:
                send += 1

            string = content[stringoffset:send]
            if string == b"":
                empty_len = len(self._emptystring)
                for j in range(i, stringcount):
                    j += 1
                    nextoffset = b2i(content, j * size)
                    nextend = nextoffset
                    while content[nextend] != 0:
                        nextend += 1
                    nextstring = content[nextoffset:nextend]
                    if nextstring != b"":
                        self._emptystring.append((i - 1, nextstring))
                        break
                if len(self._emptystring) == empty_len:
                    self._emptystring.append(i - 1)
            self.AddString(string)
        self._loaded = True

    def load_tbl_with_chk(
        self,
        content: bytes,
        init_chkt: tuple[CHK, StringIdMap, StringIdMap, StringIdMap],
    ) -> None:
        # ut.ep_assert(not self._loaded, "String data are already loaded")
        self._datatb.clear()
        self._stringmap.clear()
        self._dataindextb.clear()
        self._emptystring.clear()
        self._first_extended_string = None
        self._capacity = self._saveentry

        chkt, unitmap, locmap, swmap = init_chkt

        unix = chkt.getsection("UNIx")
        try:
            swnm = chkt.getsection("SWNM")
        except KeyError:
            swnm = None
        sprp = chkt.getsection("SPRP")
        forc = chkt.getsection("FORC")

        locdict = dict()
        swnmdict = dict()
        unitdict = dict()

        # Get location names
        try:
            mrgn = chkt.getsection("MRGN")
        except KeyError:
            pass  # Not Required for Melee
        else:
            if mrgn:
                locn = len(mrgn) // 20
                for i in range(locn):
                    locstrid = ut.b2i2(mrgn, 20 * i + 16)
                    if locstrid:
                        locdict[locstrid] = i

        # Get switch names
        if swnm:
            for i in range(256):
                swstrid = ut.b2i4(swnm, 4 * i)
                if swstrid:
                    swnmdict[swstrid] = i

        # Get unit names
        if unix:
            for i in range(228):
                unitstrid = ut.b2i2(unix, 3192 + i * 2)
                if unitstrid:
                    unitdict[unitstrid] = i

        reserved_str = set(unitdict.keys())

        if sprp:
            for i in range(2):
                mapstrid = ut.b2i2(sprp, i * 2)
                if mapstrid:
                    reserved_str.add(mapstrid)

        if forc:
            for i in range(4):
                forcstrid = ut.b2i2(forc, 8 + i * 2)
                if forcstrid:
                    reserved_str.add(forcstrid)

        size = self._loadentry
        b2i = _b2in(size)
        removed_str = set(locdict.keys()).union(swnmdict.keys()) - reserved_str
        stringcount = b2i(content, 0)

        for i in range(stringcount):
            i += 1
            stringoffset = b2i(content, i * size)
            send = stringoffset
            try:
                while content[send] != 0:
                    send += 1
            except IndexError:
                if stringoffset == send:
                    # invalid string offset
                    continue
                else:
                    # no null terminator
                    pass

            string = content[stringoffset:send]
            if string == b"" or i in removed_str:
                empty_len = len(self._emptystring)
                for j in range(i, stringcount):
                    j += 1
                    if j in removed_str:
                        continue
                    nextoffset = b2i(content, j * size)
                    nextend = nextoffset
                    while content[nextend] != 0:
                        nextend += 1
                    nextstring = content[nextoffset:nextend]
                    if nextstring != b"":
                        if j in unitdict and unit_name_encoding:
                            try:
                                nextstring = (
                                    nextstring.decode(unit_name_encoding)
                                ).encode("utf-8")
                            except UnicodeDecodeError:
                                pass
                        self._emptystring.append((i - 1, nextstring))
                        break
                if len(self._emptystring) == empty_len:
                    self._emptystring.append(i - 1)

            if string:
                if i in unitdict:
                    if unit_name_encoding:
                        try:
                            string = (string.decode(unit_name_encoding)).encode(
                                "utf-8"
                            )
                        except UnicodeDecodeError:
                            pass
                    unitmap.add_item(string, unitdict[i])
                    color_erased = _ignore_color(string)
                    if color_erased:
                        unitmap.add_item(color_erased, unitdict[i])
                if i in locdict:
                    locmap.add_item(string, locdict[i])
                if i in swnmdict:
                    swmap.add_item(string, swnmdict[i])

            if i in removed_str:
                string = b""

            self.AddString(string)
        self._loaded = True

    def AddString(self, string: str | bytes) -> int:  # noqa: N802
        ut.ep_assert(
            not self._finalized, _("Can't add new string after finalization")
        )
        # Starcraft: Remastered uses both utf-8 and multibyte encoding.
        if isinstance(string, str):
            try:
                string = ut.u2utf8(string)
            except UnicodeEncodeError:
                string = ut.u2b(string)
        if not isinstance(string, bytes):
            raise ut.EPError(_("Invalid type for string") + f": {string}")

        stringindex = len(self._dataindextb)

        # If duplicate text exist -> just proxy it
        try:
            repr_stringid = self._stringmap[string]
            dataindex = self._dataindextb[repr_stringid]
            self._dataindextb.append(dataindex)
            self._capacity += self._saveentry  # just string offset

        # Else -> Create new entry
        except KeyError:
            if self._emptystring and self._loaded and self._first_extended_string:
                emptystring = self._emptystring.pop(0)
                if isinstance(emptystring, tuple):
                    stringindex, nextstring = emptystring
                else:
                    stringindex = emptystring
                    nextstring = self._first_extended_string
                dataindex = self._datatb.index(nextstring)
                self._datatb.insert(dataindex, string)
                for i, v in enumerate(self._dataindextb):
                    if v >= dataindex:
                        self._dataindextb[i] += 1
                self._dataindextb[stringindex] = dataindex
                # string + b'\0'
                self._capacity += _roundup_by_4(len(string) + 1)
            else:
                if self._first_extended_string is None:
                    self._first_extended_string = string
                dataindex = len(self._datatb)
                self._datatb.append(string)
                self._dataindextb.append(dataindex)
                # string + b'\0' + string offset
                self._capacity += _roundup_by_4(len(string) + 1) + self._saveentry
            self._stringmap[string] = stringindex

        ut.ep_assert(
            self._capacity < (1 << (8 * self._saveentry)),
            _("String table overflow"),
        )
        ut.ep_assert(0 <= stringindex < 65535, _("String count exceeded"))

        return stringindex

    def GetString(self, index: int) -> bytes | None:  # noqa: N802
        if index == 0:
            return None
        else:
            try:
                dataindex = self._dataindextb[index - 1]
                return self._datatb[dataindex]
            except IndexError:
                return None

    def GetStringIndex(self, string: str | bytes) -> int:  # noqa: N802
        try:
            return self._stringmap[ut.u2b(string)] + 1
        except KeyError:
            try:
                return self._stringmap[ut.u2utf8(string)] + 1
            except KeyError:
                return self.AddString(string) + 1

    def save_tbl(self) -> bytes:
        if self._finalized:
            return self._tbldata

        # calculate offset of each string
        outbytes = []
        self._stroffset.clear()
        size = self._saveentry
        outindex = _roundup_by_4(size * len(self._dataindextb) + size)

        for s in self._datatb:
            self._stroffset.append(outindex)
            outindex += _roundup_by_4(len(s) + 1)
        i2b = _i2bn(size)

        # String count
        outbytes.append(i2b(len(self._dataindextb)))

        # String offsets
        for dataidx in self._dataindextb:
            outbytes.append(i2b(self._stroffset[dataidx]))

        tablesize = size * (len(self._dataindextb) + 1)
        for _n in range(_roundup_by_4(tablesize) - tablesize):
            outbytes.append(b"\0")

        # String data
        for s in self._datatb:
            outbytes.append(s)
            for _n in range(_roundup_by_4(len(s) + 1) - len(s)):
                outbytes.append(b"\0")

        self._tbldata = b"".join(outbytes)
        return self._tbldata

    def finalize(self) -> tuple[bytes, list[int]]:
        self.save_tbl()
        self._finalized = True
        return self._tbldata, self._stroffset

    def ForceAddString(self, string: str | bytes) -> int:  # noqa: N802
        ut.ep_assert(
            not self._finalized, _("Can't add new string after finalization")
        )
        string = ut.u2b(string)  # Starcraft uses multibyte encoding.
        if not isinstance(string, bytes):
            raise ut.EPError(_("Invalid type for string"))

        stringindex = len(self._dataindextb)

        # Always -> Create new entry
        dataindex = len(self._datatb)
        # self._stringmap[string] = stringindex
        self._datatb.append(string)
        self._dataindextb.append(dataindex)
        # string + b'\0' + string offset
        self._capacity += _roundup_by_4(len(string) + 1) + self._saveentry

        ut.ep_assert(
            self._capacity < (1 << (8 * self._saveentry)),
            _("String table overflow"),
        )
        ut.ep_assert(0 <= stringindex < 65535, _("String count exceeded"))

        return stringindex



# Copyright 2018 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import functools
from collections.abc import Callable

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..core.curpl import GetCPCache
from ..core.eudfunc.eudf import _EUDPredefineParam, _EUDPredefineReturn
from ..core.eudfunc.eudtypedfuncn import EUDTypedFuncN
from . import modcurpl as cp
from . import readtable

cpcache = GetCPCache()


@functools.cache
def _read_epd_func(
    mask: int,
    initvals: tuple[int, ...],
    *args: tuple[int, ...],
    _check_empty: bool = False,
) -> EUDTypedFuncN:
    @_EUDPredefineReturn(len(args))
    @_EUDPredefineParam(cp.CP)
    @c.EUDFunc
    def readerfunc(targetplayer):
        ret = readerfunc._frets
        done = c.Forward()

        if _check_empty:
            check = c.Forward()
            check << c.RawTrigger(
                conditions=c.Deaths(cp.CP, c.Exactly, 0, 0),
                actions=[retv.SetNumber(0) for retv in ret]
                + [c.SetNextPtr(check, done)],
            )
            init = c.NextTrigger()

        cs.DoActions(
            [retv.SetNumber(initval) for retv, initval in zip(ret, initvals)]
        )

        for nth, i in enumerate(ut.bits(mask)):
            if all(arg[nth] == 0 for arg in args):
                continue
            c.RawTrigger(
                conditions=c.DeathsX(cp.CP, c.AtLeast, 1, 0, i),
                actions=[
                    retv.AddNumber(arg[nth])
                    for retv, arg in zip(ret, args)
                    if arg[nth] != 0
                ],
            )

        done << c.NextTrigger()
        cp.f_setcurpl2cpcache(
            actions=c.SetNextPtr(check, init) if _check_empty else []
        )
        # return ut.List2Assignable(ret)

    return readerfunc


@functools.cache
def _readtable_epd_func(mask: int, shift: int) -> EUDTypedFuncN:
    @_EUDPredefineReturn(1)
    @_EUDPredefineParam(cp.CP)
    @c.EUDFunc
    def readerfunc(targetplayer):
        ret = readerfunc._frets[0]
        readtrg = readtable._insert_or_get(mask, shift)
        nexttrg = c.Forward()
        c.RawTrigger(
            nextptr=readtrg,
            actions=[
                c.SetNextPtr(readtable.read_end_common, cpcache.GetVTable()),
                c.SetNextPtr(cpcache.GetVTable(), nexttrg),
                cpcache.SetDest(ut.EPD(0x6509B0)),
                c.SetMemory(
                    readtable.copy_ret + 16, c.SetTo, ut.EPD(ret.getValueAddr())
                ),
            ],
        )
        nexttrg << c.NextTrigger()
        # return ret

    return readerfunc


def f_readgen_epd(
    mask: int,
    *args: tuple[int, Callable[[int], int]],
    docstring: str | None = None,
    _check_empty: bool = False,
) -> EUDTypedFuncN:
    mask = mask & 0xFFFFFFFF
    initvals = tuple(arg[0] for arg in args)
    vals = tuple(tuple(arg[1](i) for i in ut.bits(mask)) for arg in args)

    if (
        _check_empty is False
        and len(initvals) == 1
        and initvals[0] == 0
        and readtable._is_consecutive(mask)
    ):
        maybe_shift = vals[0]
        bits = ut.bits(mask)
        bit0 = bits.__next__()
        shift0 = maybe_shift[0]
        can_be_lshift = bit0 < shift0
        if can_be_lshift:
            coefficient, rem = divmod(shift0, bit0)
        else:
            coefficient, rem = divmod(bit0, shift0)
        if rem == 0 and coefficient & (coefficient - 1) == 0 and coefficient != 0:
            is_readtable = 1
            for bit, shift in zip(bits, maybe_shift[1:]):
                if can_be_lshift:
                    if bit * coefficient != shift:
                        break
                elif shift * coefficient != bit:
                    break
                is_readtable += 1
            if is_readtable == len(maybe_shift):
                signed_shift = coefficient.bit_length() - 1
                if not can_be_lshift:
                    signed_shift = -signed_shift
                readerfunc = _readtable_epd_func(mask, signed_shift)
                if docstring:
                    readerfunc.__doc__ = docstring
                return readerfunc

    readerfunc = _read_epd_func(mask, initvals, *vals, _check_empty=_check_empty)
    if docstring:
        readerfunc.__doc__ = docstring
    return readerfunc


@functools.cache
def _read_cp_func(
    mask: int,
    initvals: tuple[int, ...],
    *args: tuple[int, ...],
    _check_empty: bool = False,
) -> Callable:
    @_EUDPredefineReturn(len(args))
    @c.EUDFunc
    def reader():
        ret = reader._frets
        init_actions = [
            retv.SetNumber(initval) for retv, initval in zip(ret, initvals)
        ]
        if _check_empty:
            check, read_start = c.Forward(), c.Forward()
            init_actions.append(c.SetNextPtr(check, read_start))
        cs.DoActions(init_actions)

        if _check_empty:
            done = c.Forward()
            check << c.RawTrigger(
                conditions=c.Deaths(cp.CP, c.Exactly, 0, 0),
                actions=[
                    retv.SetNumber(0)
                    for retv, initval in zip(ret, initvals)
                    if initval != 0
                ]
                + [c.SetNextPtr(check, done)],
            )
            read_start << c.NextTrigger()

        for nth, i in enumerate(ut.bits(mask)):
            if all(arg[nth] == 0 for arg in args):
                continue
            c.RawTrigger(
                conditions=c.DeathsX(cp.CP, c.AtLeast, 1, 0, i),
                actions=[
                    retv.AddNumber(arg[nth])
                    for retv, arg in zip(ret, args)
                    if arg[nth] != 0
                ],
            )

        if _check_empty:
            done << c.NextTrigger()
        # return ut.List2Assignable(ret)

    def readerfunc(cpo, **kwargs):
        if not isinstance(cpo, int) or cpo != 0:
            cs.DoActions(c.SetMemory(0x6509B0, c.Add, cpo))
        ret = [reader(**kwargs)]
        if not isinstance(cpo, int) or cpo != 0:
            cs.DoActions(c.SetMemory(0x6509B0, c.Add, -cpo))
        return ut.List2Assignable(ret)

    return readerfunc


def f_readgen_cp(
    mask: int,
    *args: tuple[int, Callable[[int], int]],
    docstring: str | None = None,
    _check_empty: bool = False,
) -> Callable:
    mask = mask & 0xFFFFFFFF
    initvals = tuple(arg[0] for arg in args)
    vals = tuple(tuple(arg[1](i) for i in ut.bits(mask)) for arg in args)

    if (
        _check_empty is False
        and len(initvals) == 1
        and initvals[0] == 0
        and readtable._is_consecutive(mask)
    ):
        maybe_shift = vals[0]
        bits = ut.bits(mask)
        bit0 = bits.__next__()
        shift0 = maybe_shift[0]
        can_be_lshift = bit0 < shift0
        if can_be_lshift:
            coefficient, rem = divmod(shift0, bit0)
        else:
            coefficient, rem = divmod(bit0, shift0)
        if rem == 0 and coefficient & (coefficient - 1) == 0 and coefficient != 0:
            is_readtable = 1
            for bit, shift in zip(bits, maybe_shift[1:]):
                if can_be_lshift:
                    if bit * coefficient != shift:
                        break
                elif shift * coefficient != bit:
                    break
                is_readtable += 1
            if is_readtable == len(maybe_shift):
                signed_shift = coefficient.bit_length() - 1
                if not can_be_lshift:
                    signed_shift = -signed_shift
                readerfunc = readtable._cp_caller(
                    readtable._insert_or_get(mask, signed_shift)
                )
                if docstring:
                    readerfunc.__doc__ = docstring
                return readerfunc

    readerfunc = _read_cp_func(mask, initvals, *vals, _check_empty=_check_empty)
    if docstring:
        readerfunc.__doc__ = docstring
    return readerfunc


def f_maskread_epd(targetplayer, mask, *, _fdict={}, **kwargs):
    if readtable._is_consecutive(mask):
        return readtable._epd_caller(readtable._insert_or_get(mask, 0))(
            targetplayer, **kwargs
        )

    if mask not in _fdict:
        _fdict[mask] = f_readgen_epd(mask, (0, lambda x: x))
    return _fdict[mask](targetplayer, **kwargs)


def f_maskread_cp(cpo, mask, *, _fdict={}, **kwargs):
    if readtable._is_consecutive(mask):
        return readtable._cp_caller(readtable._insert_or_get(mask, 0))(cpo, **kwargs)

    if mask not in _fdict:
        _fdict[mask] = f_readgen_cp(mask, (0, lambda x: x))
    return _fdict[mask](cpo, **kwargs)



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..core.eudfunc import EUDFullFunc
from ..core.mapdata.stringmap import ForceAddString
from ..eudlib.utilf.userpl import IsUserCP, f_getuserplayerid
from ..localize import _
from ..memio import f_getcurpl, f_setcurpl
from ..scdata import CurrentPlayer
from .cpprint import FixedText, f_cpstr_print, f_gettextptr
from .cpstr import GetMapStringAddr
from .fmtprint import _format_args
from .strfunc import f_strlen_epd
from .texteffect import (
    TextFX_FadeIn,
    TextFX_FadeOut,
    TextFX_Remove,
    get_textfx_timer,
)

_display_text = c.DisplayText(0)


@EUDFullFunc(
    [
        (ut.EPD(0x640B58), c.Add, 0),
        (ut.EPD(_display_text) + 1, c.SetTo, 0),
    ],
    [None, c.TrgString],
)
def DisplayTextAt(line, text):  # noqa: N802
    with FixedText(_display_text):
        c.VProc([line, text], [])
        c.RawTrigger(
            conditions=c.Memory(0x640B58, c.AtLeast, 11),
            actions=c.SetMemory(0x640B58, c.Subtract, 11),
        )


def _tag_print(identifier, *args, encoding="UTF-8"):
    args = ut.FlattenList(args)
    ut.ep_assert(len(args) > 0, "No text provided for tagprint")

    try:
        args[0] = identifier + args[0]
    except TypeError:
        try:
            args[0] = identifier.decode("ascii") + args[0]
        except TypeError:
            args.insert(0, identifier)

    for arg in args:
        if isinstance(arg, bytes):
            arg = arg.decode(encoding)
        if isinstance(arg, str) and "\n" in arg:
            new_arg = arg.split("\n")[0]
            for line in arg.split("\n")[1:]:
                if line == "":
                    new_arg += "\n"
                else:
                    new_arg += "\n" + identifier.decode("ascii") + line
            f_cpstr_print(new_arg, EOS=False, encoding=encoding)
        else:
            f_cpstr_print(arg, EOS=False, encoding=encoding)

    cs.DoActions(c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0))


class StringBuffer(c.EUDStruct):
    """Object for storing single modifiable string.

    Manipulating STR section is easy. :)
    You can do anything you would do with normal string with StringBuffer.
    """

    _fields_ = (
        ("StringIndex", c.TrgString),  # static
        "epd",  # static
        "pos",
        "capacity",  # static
    )

    _method_template = c.Forward()
    _cpbranch = c.Forward()

    def __init__(self, content=None, *, _from=None):
        """Constructor for StringBuffer

        :param content: Initial StringBuffer content / capacity. Capacity of
            StringBuffer is determined by size of this. If content is integer,
            then initial capacity and content of StringBuffer will be set to
            content(int) and empty string.

        :type content: str, bytes, int
        """
        if _from is None or (
            isinstance(_from, StringBuffer) and c.IsConstExpr(_from)
        ):
            if _from is None:
                if content is None:
                    content = "\r" * 218
                elif isinstance(content, int):
                    content = "\r" * content
                else:
                    content = ut.u2utf8(content)

                string_id = ForceAddString(content)
                epd = ut.EPD(GetMapStringAddr(string_id))
                capacity = len(content)
                super().__init__(
                    _from=None, _static_initval=[string_id, epd, 0, capacity]
                )
            else:
                string_id = _from._StringIndex
                epd = _from._epd
                capacity = _from._capacity
                super().__init__(_from=_from)

            # static fields
            object.__setattr__(self, "_initialized", False)
            self._StringIndex = string_id
            self._epd = epd
            self._capacity = capacity
            self._initialized = True

        else:
            super().__init__(_from=_from)

    def constructor(self, *args, **kwargs):
        raise ut.EPError(_("Dynamically allocating StringBuffer is not supported"))

    def constructor_static(self, *args, **kwargs):
        pass

    @classmethod
    def alloc(cls, *args, **kwargs):
        raise ut.EPError(_("Dynamically allocating StringBuffer is not supported"))

    @classmethod
    def free(cls, data):
        raise ut.EPError(_("Dynamically allocating StringBuffer is not supported"))

    def destructor(self):
        raise ut.EPError(_("Dynamically allocating StringBuffer is not supported"))

    # Initializer

    def setall(self, values):
        raise ut.EPError(_("Can't mutate field of StringBuffer except 'pos'"))

    def copyto(self, inst):
        raise ut.EPError(_("Can't clone StringBuffer, which is reference type."))

    # Field setter & getter

    def getfield(self, name):
        # get static field
        if c.IsConstExpr(self.getValue()) and name != "pos":
            return getattr(self, f"_{name}")
        return super().getfield(name)

    def setfield(self, name, value):
        ut.ep_assert(
            name == "pos", _("Can't mutate field of StringBuffer except 'pos'")
        )
        super().setfield(name, value)

    def iaddattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().iaddattr(name, value)

    def isubtractattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().isubtractattr(name, value)

    def isubattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().isubattr(name, value)

    def imulattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().imulattr(name, value)

    def ifloordivattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().ifloordivattr(name, value)

    def imodattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().imodattr(name, value)

    def ilshiftattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().ilshiftattr(name, value)

    def irshiftattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().irshiftattr(name, value)

    def ipowattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().ipowattr(name, value)

    def iandattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().iandattr(name, value)

    def iorattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().iorattr(name, value)

    def ixorattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().ixorattr(name, value)

    def iinvertattr(self, name, value):
        if name != "pos":
            raise AttributeError
        super().iinvertattr(name, value)

    @classmethod
    def _init_template(cls):
        c.PushTriggerScope()
        cls._method_template << c.NextTrigger()
        _localcp = f_getuserplayerid()
        cls._cpbranch << c.RawTrigger(
            conditions=IsUserCP(), actions=c.SetNextPtr(cls._cpbranch, 0)
        )
        c.PopTriggerScope()

    @classmethod
    def _cpblock(cls):
        if not cls._method_template.IsSet():
            cls._init_template()
        end, ontrue = c.Forward(), c.Forward()
        c.RawTrigger(
            nextptr=cls._method_template,
            actions=[
                c.SetNextPtr(cls._cpbranch, end),
                c.SetMemory(cls._cpbranch + 348, c.SetTo, ontrue),
            ],
        )
        ontrue << c.NextTrigger()
        yield end
        end << c.NextTrigger()

    def append(self, *args):
        for _end in self._cpblock():
            f_setcurpl(self.pos)
            f_cpstr_print(*args)
            self.pos = f_getcurpl()
            f_setcurpl(f_getuserplayerid())

    def appendf(self, format_string, *args):
        self.append(*_format_args(format_string, *args))

    def insert(self, index, *args):
        for _end in self._cpblock():
            f_setcurpl(self.epd + index)
            f_cpstr_print(*args, EOS=False)
            self.pos = f_getcurpl()
            f_setcurpl(f_getuserplayerid())

    def insertf(self, index, format_string, *args):
        self.insert(index, *_format_args(format_string, *args))

    def delete(self, start, length=1):
        for _end in self._cpblock():
            pos = self.epd + start
            self.pos = pos
            f_setcurpl(pos)
            cs.DoActions(
                [
                    c.SetDeaths(CurrentPlayer, c.SetTo, 0x0D0D0D0D, 0),
                    c.AddCurrentPlayer(1),
                ]
                for _ in range(length)
            )
            f_setcurpl(f_getuserplayerid())

    def Display(self):  # noqa: N802
        cs.DoActions(c.DisplayText(self.StringIndex))

    def DisplayAt(self, line, _f={}):  # noqa: N802
        for _end in self._cpblock():
            with FixedText(c.DisplayText(self.StringIndex)):
                if c.IsEUDVariable(line):
                    c.VProc(line, line.QueueAddTo(ut.EPD(0x640B58)))
                elif not isinstance(line, int) or line != 0:
                    cs.DoActions(c.SetMemory(0x640B58, c.Add, line))
                if not isinstance(line, int) or line != 0:
                    c.RawTrigger(
                        conditions=c.Memory(0x640B58, c.AtLeast, 11),
                        actions=c.SetMemory(0x640B58, c.Subtract, 11),
                    )

    def print(self, *args):
        if len(args) == 1 and isinstance(args[0], str):
            return cs.DoActions(c.DisplayText(args[0]))
        for _end in self._cpblock():
            f_setcurpl(self.epd)
            f_cpstr_print(*args, EOS=False)
            cs.DoActions(
                c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0),
                *c.SetCurrentPlayer(f_getuserplayerid()),
                c.DisplayText(self.StringIndex),
            )

    def printf(self, format_string, *args):
        self.print(*_format_args(format_string, *args))

    def printAt(self, line, *args):  # noqa: N802
        if len(args) == 1 and isinstance(args[0], str):
            return DisplayTextAt(line, args[0])
        for _end in self._cpblock():
            f_setcurpl(self.epd)
            f_cpstr_print(*args, EOS=False)
            cs.DoActions(
                c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0),
                *c.SetCurrentPlayer(f_getuserplayerid()),
            )
            self.DisplayAt(line)

    def printfAt(self, line, format_string, *args):  # noqa: N802
        self.printAt(line, *_format_args(format_string, *args))

    def _display_at(self, prevpos, line):
        f_setcurpl(f_getuserplayerid())
        if isinstance(line, int):
            if line == 10:
                prevptr = f_gettextptr()
                prevptr -= 1
                if cs.EUDIf()(prevpos == prevptr):
                    self.DisplayAt(10)
                if cs.EUDElse()():
                    self.Display()
                cs.EUDEndIf()
            elif 0 <= line < 10:
                self.DisplayAt(line)
            elif line >= -11:
                if cs.EUDIf()(prevpos == -1):
                    if line == -1:
                        self.Display()
                    else:
                        self.DisplayAt(11 + line)
                if cs.EUDElse()():
                    with FixedText(c.DisplayText(self.StringIndex)):
                        c.VProc(prevpos, prevpos.SetDest(ut.EPD(0x640B58)))
                cs.EUDEndIf()
        else:
            if cs.EUDIf()(line == 10):
                prevptr = f_gettextptr()
                prevptr -= 1
                if cs.EUDIf()(prevpos == prevptr):
                    self.DisplayAt(10)
                if cs.EUDElse()():
                    self.Display()
                cs.EUDEndIf()
            if cs.EUDElseIfNot()(line >= 10):
                self.DisplayAt(line)
            if cs.EUDElseIf()(line >= 0xFFFFFFF5):
                if cs.EUDIf()(prevpos == -1):
                    if cs.EUDIf()(line == -1):
                        self.Display()
                    if cs.EUDElse()():
                        self.DisplayAt(11 + line)
                    cs.EUDEndIf()
                if cs.EUDElse()():
                    with FixedText(c.DisplayText(self.StringIndex)):
                        c.VProc(prevpos, prevpos.SetDest(ut.EPD(0x640B58)))
                cs.EUDEndIf()
            cs.EUDEndIf()

    def tagprint(self, format_string, *args, line=-1, tag=None):
        fmtargs = _format_args(format_string, *args)
        if tag is None:
            if len(fmtargs) == 1:
                tag = fmtargs[0]
            else:
                tag = fmtargs
        if not self._method_template.IsSet():
            self._init_template()
        end, ontrue = c.Forward(), c.Forward()
        ret = c.EUDVariable()
        c.RawTrigger(
            nextptr=self._method_template,
            actions=[
                c.SetNextPtr(self._cpbranch, end),
                c.SetMemory(self._cpbranch + 348, c.SetTo, ontrue),
                ret.SetNumber(-1),
            ],
        )
        ontrue << c.NextTrigger()
        prevpos = TextFX_Remove(tag)
        f_setcurpl(self.epd)
        _, _, identifier = get_textfx_timer(tag)
        _tag_print(identifier, *fmtargs)
        self._display_at(prevpos, line)
        end << c.NextTrigger()

    def Play(self):  # noqa: N802
        cs.DoActions(c.PlayWAV(self.StringIndex))

    def fadeIn(self, *args, color=None, wait=1, reset=True, line=-1, tag=None):  # noqa: N802
        if tag is None:
            if len(args) == 1:
                tag = args[0]
            else:
                tag = args
        if not self._method_template.IsSet():
            self._init_template()
        end, ontrue = c.Forward(), c.Forward()
        ret = c.EUDVariable()
        c.RawTrigger(
            nextptr=self._method_template,
            actions=[
                c.SetNextPtr(self._cpbranch, end),
                c.SetMemory(self._cpbranch + 348, c.SetTo, ontrue),
                ret.SetNumber(-1),
            ],
        )
        ontrue << c.NextTrigger()
        prevpos = TextFX_Remove(tag)
        f_setcurpl(self.epd)
        ret << TextFX_FadeIn(*args, color=color, wait=wait, reset=reset, tag=tag)
        self._display_at(prevpos, line)
        end << c.NextTrigger()
        return ret

    def fadeInf(  # noqa: N802
        self,
        format_string,
        *args,
        color=None,
        wait=1,
        reset=True,
        line=-1,
        tag=None,
    ):
        fmtargs = _format_args(format_string, *args)
        return self.fadeIn(
            *fmtargs, color=color, wait=wait, reset=reset, line=line, tag=tag
        )

    def fadeOut(  # noqa: N802
        self, *args, color=None, wait=1, reset=True, line=-1, tag=None
    ):
        if tag is None:
            if len(args) == 1:
                tag = args[0]
            else:
                tag = args
        if not self._method_template.IsSet():
            self._init_template()
        end, ontrue = c.Forward(), c.Forward()
        ret = c.EUDVariable()
        c.RawTrigger(
            nextptr=self._method_template,
            actions=[
                c.SetNextPtr(self._cpbranch, end),
                c.SetMemory(self._cpbranch + 348, c.SetTo, ontrue),
                ret.SetNumber(-1),
            ],
        )
        ontrue << c.NextTrigger()
        prevpos = TextFX_Remove(tag)
        f_setcurpl(self.epd)
        ret << TextFX_FadeOut(*args, color=color, wait=wait, reset=reset, tag=tag)
        self._display_at(prevpos, line)
        end << c.NextTrigger()
        return ret

    def fadeOutf(  # noqa: N802
        self,
        format_string,
        *args,
        color=None,
        wait=1,
        reset=True,
        line=-1,
        tag=None,
    ):
        fmtargs = _format_args(format_string, *args)
        return self.fadeOut(
            *fmtargs, color=color, wait=wait, reset=reset, line=line, tag=tag
        )

    def length(self):
        return f_strlen_epd(self.epd)


_globalsb = None


def GetGlobalStringBuffer():  # noqa: N802
    global _globalsb
    if _globalsb is None:
        _globalsb = StringBuffer(1023)
    return _globalsb


# FIXME: Optimize these functions to cache string contents
def f_simpleprint(*args, spaced=True):
    # Add spaces between arguments
    if spaced:
        spaced_args = []
        for arg in args:
            spaced_args.extend([arg, " "])
        args = spaced_args[:-1]

    # Print
    oldcp = f_getcurpl()
    f_setcurpl(f_getuserplayerid())
    gsb = GetGlobalStringBuffer()
    gsb.print(*args)
    f_setcurpl(oldcp)


def f_println(format_string, *args):
    gsb = GetGlobalStringBuffer()
    gsb.printf(format_string, *args)


def f_printAt(line, format_string, *args):  # noqa: N802
    gsb = GetGlobalStringBuffer()
    gsb.printfAt(line, format_string, *args)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .bwepdio import (
    f_badd_epd,
    f_bread_epd,
    f_bsubtract_epd,
    f_bwrite_epd,
    f_maskwrite_epd,
    f_wadd_epd,
    f_wread_epd,
    f_wsubtract_epd,
    f_wwrite_epd,
)
from .byterw import EUDByteReader, EUDByteStream, EUDByteWriter
from .cpbyterw import CPByteWriter
from .cpmemio import (
    f_bread_cp,
    f_bwrite_cp,
    f_dwadd_cp,
    f_dwepdread_cp,
    f_dwread_cp,
    f_dwsubtract_cp,
    f_dwwrite_cp,
    f_epdread_cp,
    f_maskwrite_cp,
    f_wread_cp,
    f_wwrite_cp,
)
from .dwepdio import (
    f_dwadd_epd,
    f_dwbreak,
    f_dwbreak2,
    f_dwepdread_epd,
    f_dwepdread_epd_safe,
    f_dwread_epd,
    f_dwread_epd_safe,
    f_dwsubtract_epd,
    f_dwwrite_epd,
    f_epdread_epd,
    f_epdread_epd_safe,
    f_flagread_epd,
)
from .mblockio import f_memcmp, f_memcpy, f_repmovsd_epd
from .memifgen import (
    f_maskread_cp,
    f_maskread_epd,
    f_readgen_cp,
    f_readgen_epd,
)
from .modcurpl import f_addcurpl, f_getcurpl, f_setcurpl, f_setcurpl2cpcache
from .ptrmemio import f_bread, f_bwrite, f_dwread, f_dwwrite, f_wread, f_wwrite
from .specialized import (
    f_cunitepdread_cp,
    f_cunitepdread_epd,
    f_cunitread_cp,
    f_cunitread_epd,
    f_epdcunitread_cp,
    f_epdcunitread_epd,
    f_epdspriteread_cp,
    f_epdspriteread_epd,
    f_posread_cp,
    f_posread_epd,
    f_spriteepdread_cp,
    f_spriteepdread_epd,
    f_spriteread_cp,
    f_spriteread_epd,
)
from .varrayreader import EUDVArrayReader

__all__ = [
    "f_badd_epd",
    "f_bread_epd",
    "f_bsubtract_epd",
    "f_bwrite_epd",
    "f_maskwrite_epd",
    "f_wadd_epd",
    "f_wread_epd",
    "f_wsubtract_epd",
    "f_wwrite_epd",
    "EUDByteReader",
    "EUDByteStream",
    "EUDByteWriter",
    "CPByteWriter",
    "f_bread_cp",
    "f_bwrite_cp",
    "f_dwadd_cp",
    "f_dwepdread_cp",
    "f_dwread_cp",
    "f_dwsubtract_cp",
    "f_dwwrite_cp",
    "f_epdread_cp",
    "f_maskwrite_cp",
    "f_wread_cp",
    "f_wwrite_cp",
    "f_dwadd_epd",
    "f_dwbreak",
    "f_dwbreak2",
    "f_dwepdread_epd",
    "f_dwepdread_epd_safe",
    "f_dwread_epd",
    "f_dwread_epd_safe",
    "f_dwsubtract_epd",
    "f_dwwrite_epd",
    "f_epdread_epd",
    "f_epdread_epd_safe",
    "f_flagread_epd",
    "f_memcmp",
    "f_memcpy",
    "f_repmovsd_epd",
    "f_cunitepdread_cp",
    "f_cunitepdread_epd",
    "f_cunitread_cp",
    "f_cunitread_epd",
    "f_epdcunitread_cp",
    "f_epdcunitread_epd",
    "f_epdspriteread_cp",
    "f_epdspriteread_epd",
    "f_maskread_cp",
    "f_maskread_epd",
    "f_posread_cp",
    "f_posread_epd",
    "f_readgen_cp",
    "f_readgen_epd",
    "f_spriteepdread_cp",
    "f_spriteepdread_epd",
    "f_spriteread_cp",
    "f_spriteread_epd",
    "f_addcurpl",
    "f_getcurpl",
    "f_setcurpl",
    "f_setcurpl2cpcache",
    "f_bread",
    "f_bwrite",
    "f_dwread",
    "f_dwwrite",
    "f_wread",
    "f_wwrite",
    "EUDVArrayReader",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

import functools
from collections.abc import Iterator, Sequence
from itertools import chain
from math import log2
from typing import ClassVar, NoReturn

from typing_extensions import Self

from ...localize import _
from ...utils import EPD, EPError, EUDPeekBlock, ExprProxy, ep_assert, unProxy
from .. import rawtrigger as bt
from ..allocator import ConstExpr, Forward, IsConstExpr
from ..curpl import GetCPCache
from ..inplacecw import iand, ilshift, ior, irshift, iset, isub, ixor
from ..variable import (
    EUDLightVariable,
    EUDVariable,
    IsEUDVariable,
    SeqCompute,
    SetVariables,
    VProc,
)
from ..variable.eudv import process_dest
from ..variable.vbuf import get_current_custom_varbuffer

cpcache: EUDVariable = GetCPCache()


class _EUDVArrayData(ConstExpr):
    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls, None)

    def __init__(
        self,
        initvar: Sequence[tuple[ConstExpr | int, ConstExpr | int, ConstExpr | int]],
    ) -> None:
        super().__init__()
        init = []
        for i, items in enumerate(initvar):
            dest, value, nextptr = items
            value = unProxy(value)
            if not isinstance(value, (int, ConstExpr)):  # noqa: UP038
                raise EPError(_("Invalid item #{}: {}").format(i, items))
            nextptr = unProxy(nextptr)
            if not isinstance(nextptr, (int, ConstExpr)):  # noqa: UP038
                raise EPError(_("Invalid item #{}: {}").format(i, items))
            init.append((0xFFFFFFFF, process_dest(dest), value, 0x072D0000, nextptr))
        self._init = init

    def Evaluate(self):  # noqa: N802
        evb = get_current_custom_varbuffer()
        if self not in evb._vdict:
            evb.create_vartriggers(self, self._init)

        return evb._vdict[self].Evaluate()


bt.PushTriggerScope()
_getter = bt.RawTrigger(nextptr=0, actions=bt.SetMemoryEPD(0, bt.SetTo, 0))
_getter_dst: ConstExpr = EPD(_getter) + 86  # this and nextptr are set by __getitem__
# this is set by inner triggers of _EUDVArray
_getter_val: ConstExpr = EPD(_getter) + 87

_lv = EUDLightVariable()
_convert_index = bt.SetMemoryEPD(0, bt.Add, 0)
_get_index = bt.RawTrigger(
    nextptr=0,
    actions=[_convert_index, bt.SetMemory(_convert_index + 20, bt.SetTo, 0)],
)
_index_for_get: list[bt.RawTrigger] = []
for i in range(28):
    _index_for_get.append(
        bt.RawTrigger(
            nextptr=_get_index if i == 0 else _index_for_get[i - 1],
            conditions=_lv.AtLeastX(1, 1 << i),
            actions=bt.SetMemory(_convert_index + 20, bt.Add, 72 << i),
        )
    )

_convert_index = bt.SetMemoryEPD(0, bt.SetTo, 0)
_set_index = bt.RawTrigger(nextptr=0, actions=_convert_index)
_index_for_set: list[bt.RawTrigger] = []
for i in range(28):
    _index_for_set.append(
        bt.RawTrigger(
            nextptr=_set_index if i == 0 else _index_for_set[i - 1],
            conditions=_lv.AtLeastX(1, 1 << i),
            actions=bt.SetMemory(_convert_index + 16, bt.Add, 18 << i),
        )
    )
bt.PopTriggerScope()


class _EUDVArray(ExprProxy):
    dont_flatten = True

    def __init__(
        self,
        initvar: Sequence[
            ConstExpr
            | int
            | tuple[ConstExpr | int, ConstExpr | int, ConstExpr | int]
        ]
        | None = None,
        *,
        _from=None,
        _size: int,
        _basetype: type | None,
    ) -> None:
        ep_assert(0 <= _size < 2**28, f"Invalid EUDVArray capacity: {_size}")
        self._size = _size
        self._basetype = _basetype

        # Initialization from value
        if _from is not None:
            baseobj = _from

        else:
            init: Sequence[tuple[ConstExpr | int, ConstExpr | int, ConstExpr | int]]
            if initvar is None:
                init = [(_getter_val, 0, _getter)] * _size
            else:
                init = [
                    x if isinstance(x, tuple) else (_getter_val, x, _getter)
                    for x in initvar
                ]

            # For python iterables
            baseobj = _EUDVArrayData(init)

        super().__init__(baseobj)
        self._epd = EPD(self)

    def _bound_check(self, index: object) -> None:
        if isinstance(index, int) and not (0 <= index < self._size):
            e = _(
                "index out of bounds: the length of EUDVArray is {}"
                " but the index is {}"
            )
            raise EPError(e.format(self._size, index))

    def __getitem__(self, i):
        return self.get(i)

    def __setitem__(self, i, value) -> None:
        self.set(i, value)

    def __iter__(self):
        yield from _VArrayIterator(self)

    def get(self, i, **kwargs):
        ret = kwargs["ret"][0] if "ret" in kwargs else EUDVariable()
        try:
            dst = EPD(ret.getValueAddr())
        except AttributeError:
            dst = ret

        index = unProxy(i)
        self._bound_check(index)
        nexttrg = Forward()

        if isinstance(index, EUDVariable):
            size_msb = self._size.bit_length() - 1
            convert_start = _index_for_get[size_msb]
            _index_dst = _get_index + 344
            if IsEUDVariable(self._value):
                ep_assert(
                    self._value is not index, "EUDVArray.ptr and index are alias"
                )
                # index copy to lv -> self.ptr -> _convert_index -> self
                bt.RawTrigger(
                    nextptr=index.GetVTable(),
                    actions=[
                        bt.SetNextPtr(index.GetVTable(), self._value.GetVTable()),
                        index.QueueAssignTo(_lv),
                        bt.SetNextPtr(self._value.GetVTable(), convert_start),
                        self._value.QueueAssignTo(EPD(_get_index) + 1),
                        bt.SetMemory(_index_dst, bt.SetTo, EPD(_get_index) + 1),
                        bt.SetNextPtr(_getter, nexttrg),
                        bt.SetMemoryEPD(_getter_dst, bt.SetTo, dst),
                    ],
                )
            else:
                # index copy to lv -> _convert_index -> self
                bt.RawTrigger(
                    nextptr=index.GetVTable(),
                    actions=[
                        bt.SetNextPtr(index.GetVTable(), convert_start),
                        index.QueueAssignTo(_lv),
                        bt.SetNextPtr(_get_index, self._value),
                        bt.SetMemory(_index_dst, bt.SetTo, EPD(_get_index) + 1),
                        bt.SetNextPtr(_getter, nexttrg),
                        bt.SetMemoryEPD(_getter_dst, bt.SetTo, dst),
                    ],
                )

        else:
            if IsEUDVariable(self._value):
                bt.RawTrigger(
                    nextptr=self._value.GetVTable(),
                    actions=[
                        bt.SetNextPtr(self._value.GetVTable(), 72 * index),
                        self._value.QueueAddTo(EPD(self._value.GetVTable()) + 1),
                        bt.SetNextPtr(_getter, nexttrg),
                        bt.SetMemoryEPD(_getter_dst, bt.SetTo, dst),
                    ],
                )
            else:
                bt.RawTrigger(
                    nextptr=self._value + 72 * index,
                    actions=[
                        bt.SetNextPtr(_getter, nexttrg),
                        bt.SetMemoryEPD(_getter_dst, bt.SetTo, dst),
                    ],
                )

        nexttrg << bt.NextTrigger()
        if self._basetype:
            ret = self._basetype.cast(ret)
        return ret

    def set(self, i, value, **kwargs):
        if self._basetype:
            value = self._basetype.cast(value)

        index = unProxy(i)
        value = unProxy(value)
        self._bound_check(index)

        if all(not isinstance(v, EUDVariable) for v in (self._epd, index, value)):
            bt.RawTrigger(
                actions=bt.SetMemoryEPD(87 + self._epd + 18 * index, bt.SetTo, value)
            )
            return

        nexttrg = Forward()
        size_msb = self._size.bit_length() - 1
        convert_start = _index_for_set[size_msb]

        # index and name alias
        if isinstance(index, EUDVariable) and index is value:
            if isinstance(self._epd, EUDVariable):
                VProc(
                    [self._epd, index],
                    [
                        bt.SetMemory(_set_index + 344, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(_set_index + 344)),
                        *index.QueueAssignTo(EPD(_set_index) + 87),
                    ],
                )
            else:
                VProc(
                    index,
                    [
                        bt.SetMemory(_set_index + 344, bt.SetTo, 87 + self._epd),
                        *index.QueueAssignTo(EPD(_set_index) + 87),
                    ],
                )
            bt.RawTrigger(
                nextptr=index.GetVTable(),
                actions=[
                    bt.SetNextPtr(index.GetVTable(), convert_start),
                    bt.SetNextPtr(_set_index, nexttrg),
                    *index.QueueAssignTo(_lv),
                ],
            )
            nexttrg << bt.NextTrigger()
            return

        # execution order: self._epd -> value -> index(->_convert_index)
        vv = []
        first_next_triggger = None

        writing_trigger = _set_index
        write_dest = _set_index + 344
        if isinstance(value, EUDVariable) and not isinstance(index, EUDVariable):
            writing_trigger = value.GetVTable()
            write_dest = value.getDestAddr()

        actions = [bt.SetNextPtr(writing_trigger, nexttrg)]

        for v in (self._epd, value, index):
            if isinstance(v, EUDVariable):
                vv.append(v)
                if first_next_triggger is None:
                    first_next_triggger = v.GetVTable()

        # nextptr
        for i in range(len(vv) - 1):
            actions.append(bt.SetNextPtr(vv[i].GetVTable(), vv[i + 1].GetVTable()))

        # dest
        const_dest = [87]
        if not isinstance(self._epd, EUDVariable):
            const_dest.append(self._epd)
        if not isinstance(index, EUDVariable):
            const_dest.append(18 * index)
        actions.append(bt.SetMemory(write_dest, bt.SetTo, sum(const_dest)))

        if isinstance(self._epd, EUDVariable):
            ep_assert(
                self._epd is not value,
                "EUDVArray._epd and __setitem__ value are aliased",
            )
            if len(vv) == 1:
                actions.append(bt.SetNextPtr(self._epd.GetVTable(), writing_trigger))
            actions.extend(self._epd.QueueAddTo(EPD(write_dest)))

        if isinstance(index, EUDVariable):
            actions.append(
                (
                    bt.SetNextPtr(index.GetVTable(), convert_start),
                    *index.QueueAssignTo(_lv),
                )
            )

        # value
        if isinstance(value, EUDVariable):
            if writing_trigger is value.GetVTable():
                actions.append(value.SetModifier(bt.SetTo))
            else:
                actions.extend(value.QueueAssignTo(EPD(_set_index) + 87))
        else:
            actions.append(bt.SetMemory(_set_index + 348, bt.SetTo, value))

        bt.RawTrigger(nextptr=first_next_triggger, actions=actions)
        nexttrg << bt.NextTrigger()


class _VArrayIterator:
    def __init__(self, varray: _EUDVArray):
        self._varray = varray
        self._size = varray._size

    def __iter__(self):
        from ... import ctrlstru as cs

        counter = EUDLightVariable()
        item = EUDVariable()
        loopstart, loopbody = Forward(), Forward()
        if IsEUDVariable(self._varray):
            VProc(
                self._varray,
                [
                    counter.SetNumber(self._size),
                    self._varray.QueueAssignTo(EPD(loopstart) + 1),
                    bt.SetNextPtr(_getter, loopbody),
                    bt.SetMemoryEPD(_getter_dst, bt.SetTo, EPD(item.getValueAddr())),
                ],
            )
        else:
            bt.RawTrigger(
                actions=[
                    counter.SetNumber(self._size),
                    bt.SetNextPtr(loopstart, self._varray),
                    bt.SetNextPtr(_getter, loopbody),
                    bt.SetMemoryEPD(_getter_dst, bt.SetTo, EPD(item.getValueAddr())),
                ]
            )
        counter << self._size
        if cs.EUDWhileNot()(counter == 0):
            block = EUDPeekBlock("whileblock")[1]
            loopstart << block["loopstart"]
            loopbody << bt.NextTrigger()
            yield item
            cs.EUDSetContinuePoint()
            bt.RawTrigger(
                actions=[
                    counter.SubtractNumber(1),
                    bt.SetMemory(loopstart + 4, bt.Add, 72),
                    bt.SetNextPtr(_getter, loopbody),  # needed for nested loop
                    bt.SetMemoryEPD(_getter_dst, bt.SetTo, EPD(item.getValueAddr())),
                ]
            )
        cs.EUDEndWhile()


class EUDVArray:
    def __init__(self, size: int, basetype: type | None = None):
        self._size = size
        self._basetype = basetype

    def __call__(
        self, initvar: Sequence[int | ConstExpr] | None = None, *, _from=None
    ) -> _EUDVArray:
        return _EUDVArray(
            initvar=initvar,
            _from=_from,
            _size=self._size,
            _basetype=self._basetype,
        )

    def cast(self, _from):
        return self(_from=_from)


_index = EUDLightVariable()


class BitsTrg:
    cache: ClassVar[dict[str, dict[int | str, Forward]]] = {}

    def __init__(self, key: str) -> None:
        self._key: str = key

    def __bool__(self) -> bool:
        return self._key in BitsTrg.cache

    def __iter__(self) -> Iterator[dict[int | str, Forward]]:
        if not self:
            BitsTrg.cache[self._key] = {i: Forward() for i in range(28)}
            bt.PushTriggerScope()
            yield BitsTrg.cache[self._key]
            bt.PopTriggerScope()

    def __getitem__(self, index: int | str) -> Forward:
        return BitsTrg.cache[self._key][index]

    def __setitem__(self, index: int | str, item: Forward) -> None:
        BitsTrg.cache[self._key][index] = item


@functools.cache
def _InternalVArray(size: int, basetype: type | None = None):  # noqa: N802
    ep_assert(isinstance(size, int) and size < 2**28, "invalid size")

    def _bound_check(index: object) -> None:
        index = unProxy(unProxy)
        if not isinstance(index, int) or 0 <= index < size:
            return
        e = _(
            "index out of bounds: the length of EUDVArray is {}"
            " but the index is {}"
        )
        raise EPError(e.format(size, index))

    class _VArray(ExprProxy):
        dont_flatten = True

        def __init__(self, initvars=None, *, dest=0, nextptr=0, _from=None) -> None:
            # Initialization from value
            if _from is not None:
                baseobj = _from

            else:
                # Int -> interpret as sequence of 0s
                if initvars is None:
                    initvars = [0] * size

                # For python iterables
                baseobj = _EUDVArrayData([(dest, x, nextptr) for x in initvars])

            super().__init__(baseobj)
            self._epd = EPD(self)
            self._basetype = basetype

        def Assign(self, other) -> Self:  # noqa: N802
            if not isinstance(self._value, EUDVariable):
                raise EPError(
                    _("Can't assign {} to constant expression").format(other)
                )
            if isinstance(other, type(self)):
                SetVariables([self._value, self._epd], [other, other._epd])
            elif isinstance(other, int) and other == 0:
                SetVariables([self._value, self._epd], [0, 0])
            else:
                raise EPError(_("Can't assign {} to {}").format(other, self))
            return self

        def get(self, i, **kwargs):
            if IsEUDVariable(i):
                r = self._eudget(i, **kwargs)
            else:
                _bound_check(i)
                if IsEUDVariable(self):
                    r = self._get(i, **kwargs)
                else:
                    r = self._constget(i, **kwargs)

            if self._basetype:
                r = self._basetype.cast(r)
            return r

        def _eudget(self, i, **kwargs):
            bitstrg = BitsTrg("varrget")
            for trg in bitstrg:
                trg["end"], trg["ret"] = Forward(), Forward()
                for t in range(27, -1, -1):
                    trg[t] << bt.RawTrigger(
                        conditions=_index.AtLeastX(1, 2**t),
                        actions=[
                            bt.SetMemory(trg["end"] + 4, bt.Add, 72 * (2**t)),
                            bt.SetMemory(trg["ret"] + 16, bt.Add, 18 * (2**t)),
                            bt.SetMemory(trg["ret"] + 48, bt.Add, 18 * (2**t)),
                        ],
                    )
                trg["end"] << bt.RawTrigger(
                    nextptr=0,
                    actions=[
                        trg["ret"] << bt.SetDeaths(0, bt.SetTo, 0, 0),
                        bt.SetDeaths(0, bt.SetTo, 0, 0),
                    ],
                )

            r = kwargs["ret"][0] if "ret" in kwargs else EUDVariable()
            dst = EPD(r.getValueAddr()) if IsEUDVariable(r) else r
            bits = max((size - 1).bit_length() - 1, 0)
            nptr = Forward()
            if not IsEUDVariable(self._epd):
                bt.RawTrigger(
                    nextptr=i.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["end"] + 4, bt.SetTo, self),
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, self._epd + 86),
                        bt.SetMemory(bitstrg["ret"] + 20, bt.SetTo, dst),
                        bt.SetMemory(bitstrg["ret"] + 48, bt.SetTo, self._epd + 1),
                        bt.SetMemory(bitstrg["ret"] + 52, bt.SetTo, nptr),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                    ],
                )
            else:
                VProc(
                    [self, self._epd],
                    [
                        *self.QueueAssignTo(EPD(bitstrg["end"]) + 1),
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, 86),
                        *self._epd.QueueAddTo(EPD(bitstrg["ret"]) + 4),
                        bt.SetMemory(bitstrg["ret"] + 20, bt.SetTo, dst),
                    ],
                )
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 48, bt.SetTo, 1),
                        self._epd.SetDest(EPD(bitstrg["ret"]) + 12),
                        bt.SetMemory(bitstrg["ret"] + 52, bt.SetTo, nptr),
                        bt.SetNextPtr(self._epd.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                    ],
                )
            nptr << bt.NextTrigger()
            return r

        def _get(self, i, **kwargs):
            r = kwargs["ret"][0] if "ret" in kwargs else EUDVariable()
            dst = EPD(r.getValueAddr()) if IsEUDVariable(r) else r
            vtproc = Forward()
            nptr = Forward()
            a0, a2 = Forward(), Forward()

            VProc(
                [self, self._epd],
                [
                    bt.SetMemory(vtproc + 4, bt.SetTo, 72 * i),
                    *self.QueueAddTo(EPD(vtproc) + 1),
                    bt.SetMemory(a0 + 16, bt.SetTo, 18 * i + 344 // 4),
                    *self._epd.QueueAddTo(EPD(a0) + 4),
                ],
            )
            VProc(
                self._epd,
                [
                    a0 << bt.SetDeaths(0, bt.SetTo, dst, 0),
                    bt.SetMemory(a2 + 16, bt.SetTo, 18 * i + 1),
                    self._epd.SetDest(EPD(a2) + 4),
                ],
            )
            vtproc << bt.RawTrigger(
                nextptr=0,
                actions=[a2 << bt.SetDeaths(0, bt.SetTo, nptr, 0)],
            )

            nptr << bt.NextTrigger()
            return r

        def _constget(self, i, **kwargs):
            r = kwargs["ret"][0] if "ret" in kwargs else EUDVariable()
            dst = EPD(r.getValueAddr()) if IsEUDVariable(r) else r
            nptr = Forward()

            bt.RawTrigger(
                nextptr=self + 72 * i,
                actions=[
                    bt.SetDeaths(self._epd + 18 * i + 86, bt.SetTo, dst, 0),
                    bt.SetDeaths(self._epd + 18 * i + 1, bt.SetTo, nptr, 0),
                ],
            )
            nptr << bt.NextTrigger()
            return r

        def set(self, i, val) -> None:
            _bound_check(i)
            self._set(i, bt.SetTo, val)

        def _set(self, i, modifier, val) -> None:
            if not IsEUDVariable(i):
                iset(self._epd, 18 * i + 348 // 4, modifier, val)
                return
            modifier = bt.EncodeModifier(modifier) << 24

            bitstrg = BitsTrg("varrset")
            for trg in bitstrg:
                trg["end"], trg["ret"] = Forward(), Forward()
                for t in range(27, -1, -1):
                    trg[t] << bt.RawTrigger(
                        conditions=_index.AtLeastX(1, 2**t),
                        actions=bt.SetMemory(trg["ret"] + 16, bt.Add, 18 * (2**t)),
                    )
                trg["end"] << bt.RawTrigger(
                    nextptr=0,
                    actions=[trg["ret"] << bt.SetDeaths(0, bt.SetTo, 0, 0)],
                )

            bits = max((size - 1).bit_length() - 1, 0)
            nptr = Forward()
            if IsEUDVariable(self._epd) and IsEUDVariable(val):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(bitstrg["ret"]) + 4),
                        bt.SetNextPtr(self._epd.GetVTable(), val.GetVTable()),
                        *val.QueueAssignTo(EPD(bitstrg["ret"]) + 5),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        bt.SetMemoryX(
                            bitstrg["ret"] + 24, bt.SetTo, modifier, 0xFF << 24
                        ),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            elif IsEUDVariable(self._epd):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(bitstrg["ret"]) + 4),
                        bt.SetNextPtr(self._epd.GetVTable(), i.GetVTable()),
                        bt.SetMemory(bitstrg["ret"] + 20, bt.SetTo, val),
                        bt.SetMemoryX(
                            bitstrg["ret"] + 24, bt.SetTo, modifier, 0xFF << 24
                        ),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            elif IsEUDVariable(val):
                bt.RawTrigger(
                    nextptr=val.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, self._epd + 87),
                        *val.QueueAssignTo(EPD(bitstrg["ret"]) + 5),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        bt.SetMemoryX(
                            bitstrg["ret"] + 24, bt.SetTo, modifier, 0xFF << 24
                        ),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            else:
                bt.RawTrigger(
                    nextptr=i.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, self._epd + 87),
                        bt.SetMemory(bitstrg["ret"] + 20, bt.SetTo, val),
                        bt.SetMemoryX(
                            bitstrg["ret"] + 24, bt.SetTo, modifier, 0xFF << 24
                        ),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            nptr << bt.NextTrigger()

        def fill(self, values, *, assert_expected_values_len=None) -> None:
            if assert_expected_values_len:
                ep_assert(len(values) == assert_expected_values_len)

            SeqCompute(
                [
                    (self._epd + (18 * i + 348 // 4), bt.SetTo, value)
                    for i, value in enumerate(values)
                ]
            )

        def __getitem__(self, i):
            return self.get(i)

        def __setitem__(self, i, value) -> None:
            self.set(i, value)

        def __iter__(self) -> NoReturn:
            # FIXME: add EUDVArray iterator
            raise EPError(_("Can't iterate EUDVArray"))

        def iadditem(self, i, val) -> None:
            _bound_check(i)
            self._set(i, bt.Add, val)

        # FIXME: add operator for Subtract
        def isubtractitem(self, i, val) -> None:
            _bound_check(i)
            self._set(i, bt.Subtract, val)

        def isubitem(self, i, val) -> None:
            _bound_check(i)
            if not IsEUDVariable(val):
                self._set(i, bt.Add, -val)
                return
            if not IsEUDVariable(i):
                isub(self._epd, 18 * i + 87, val)
                return
            bitstrg = BitsTrg("varrsub")
            for trg in bitstrg:
                trg["end"], trg["ret"] = Forward(), Forward()
                for t in range(27, -1, -1):
                    trg[t] << bt.RawTrigger(
                        conditions=_index.AtLeastX(1, 2**t),
                        actions=bt.SetMemory(trg["ret"] + 16, bt.Add, 18 * (2**t)),
                    )
                trg["end"] << bt.RawTrigger(
                    nextptr=0,
                    actions=[
                        bt.SetMemoryX(trg["ret"] + 20, bt.Add, -1, 0x55555555),
                        bt.SetMemoryX(trg["ret"] + 20, bt.Add, -1, 0xAAAAAAAA),
                        bt.SetMemory(trg["ret"] + 20, bt.Add, 1),
                        trg["ret"] << bt.SetDeaths(0, bt.Add, 0, 0),
                    ],
                )

            bits = max((size - 1).bit_length() - 1, 0)
            nptr = Forward()
            if IsEUDVariable(self._epd):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(bitstrg["ret"]) + 4),
                        bt.SetNextPtr(self._epd.GetVTable(), val.GetVTable()),
                        *val.QueueAssignTo(EPD(bitstrg["ret"]) + 5),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            else:
                bt.RawTrigger(
                    nextptr=val.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, self._epd + 87),
                        *val.QueueAssignTo(EPD(bitstrg["ret"]) + 5),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            nptr << bt.NextTrigger()

        # defined when val is power of 2
        def imulitem(self, i, val) -> None:
            _bound_check(i)
            if not isinstance(val, int):
                raise AttributeError
            if val == 0:
                self.set(i, 0)
                return
            # val is power of 2
            if val & (val - 1) == 0:
                self.ilshiftitem(i, int(log2(val)))
                return
            # val is negation of power of 2
            if -val & (-val - 1) == 0:
                pass
            raise AttributeError

        # defined when val is power of 2
        def ifloordivitem(self, i, val) -> None:
            _bound_check(i)
            if not isinstance(val, int):
                raise AttributeError
            if val == 0:
                raise ZeroDivisionError
            # val is power of 2
            if val & (val - 1) == 0:
                self.irshiftitem(i, int(log2(val)))
                return
            # val is negation of power of 2
            if -val & (-val - 1) == 0:
                pass
            raise AttributeError

        # defined when val is power of 2
        def imoditem(self, i, val) -> None:
            _bound_check(i)
            if not isinstance(val, int):
                raise AttributeError
            if val == 0:
                raise ZeroDivisionError
            # val is power of 2
            if val & (val - 1) == 0:
                self.ianditem(i, val - 1)
                return
            raise AttributeError

        # FIXME: merge logic with EUDVariable and VariableBase

        def ilshiftitem(self, i, n) -> None:
            _bound_check(i)
            if not isinstance(n, int):
                raise AttributeError
            if not IsEUDVariable(i):
                ilshift(self._epd, 18 * i + 87, n)
                return
            if n == 0:
                return
            mask = (1 << (n + 1)) - 1
            bitstrg = BitsTrg(f"varrlshift{n}")
            cp = 13  # CurrentPlayer
            for trg in bitstrg:
                trg["end"] = Forward()
                for t in range(27, -1, -1):
                    trg[t] << bt.RawTrigger(
                        conditions=_index.AtLeastX(1, 2**t),
                        actions=bt.SetMemory(0x6509B0, bt.Add, 18 * (2**t)),
                    )

                def itemw(mod, value, mask):
                    return bt.SetMemoryXEPD(cp, mod, value, mask)

                trg["end"] << bt.RawTrigger(
                    nextptr=cpcache.GetVTable(),
                    actions=list(
                        chain(
                            *[
                                [
                                    itemw(bt.SetTo, 0, (mask >> 1) << (k + 1)),
                                    itemw(bt.Add, (mask >> 1) << k, mask << k),
                                ]
                                for k in reversed(range(32 - n))
                            ],
                            [
                                itemw(bt.SetTo, 0, mask >> 1),
                                cpcache.SetDest(EPD(0x6509B0)),
                            ],
                        )
                    ),
                )

            bits = max((size - 1).bit_length() - 1, 0)
            nptr = Forward()
            if IsEUDVariable(self._epd):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(0x6509B0, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(0x6509B0)),
                        bt.SetNextPtr(self._epd.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(cpcache.GetVTable(), nptr),
                    ],
                )
            else:
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(0x6509B0, bt.SetTo, self._epd + 87),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(cpcache.GetVTable(), nptr),
                    ],
                )
            nptr << bt.NextTrigger()

        def irshiftitem(self, i, n) -> None:
            _bound_check(i)
            if not isinstance(n, int):
                raise AttributeError
            if not IsEUDVariable(i):
                irshift(self._epd, 18 * i + 87, n)
                return
            if n == 0:
                return
            mask = (1 << (n + 1)) - 1
            bitstrg = BitsTrg(f"varrrshift{n}")
            cp = 13  # CurrentPlayer
            for trg in bitstrg:
                trg["end"] = Forward()
                for t in range(27, -1, -1):
                    trg[t] << bt.RawTrigger(
                        conditions=_index.AtLeastX(1, 2**t),
                        actions=bt.SetMemory(0x6509B0, bt.Add, 18 * (2**t)),
                    )

                def sub(value, mask):
                    return bt.SetMemoryXEPD(cp, bt.Subtract, value, mask)

                trg["end"] << bt.RawTrigger(
                    nextptr=cpcache.GetVTable(),
                    actions=[
                        bt.SetMemoryXEPD(cp, bt.SetTo, 0, mask >> 1),
                        cpcache.SetDest(EPD(0x6509B0)),
                    ]
                    + [sub((mask >> 1) << k, mask << k) for k in range(32 - n)],
                )

            bits = max((size - 1).bit_length() - 1, 0)
            nptr = Forward()
            if IsEUDVariable(self._epd):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(0x6509B0, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(0x6509B0)),
                        bt.SetNextPtr(self._epd.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(cpcache.GetVTable(), nptr),
                    ],
                )
            else:
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(0x6509B0, bt.SetTo, self._epd + 87),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(cpcache.GetVTable(), nptr),
                    ],
                )
            nptr << bt.NextTrigger()

        def ipowitem(self, i, val) -> None:
            _bound_check(i)
            if isinstance(val, int) and val == 1:
                return
            raise AttributeError

        def ianditem(self, i, val) -> None:
            _bound_check(i)
            if not IsEUDVariable(i):
                iand(self._epd, 18 * i + 87, val)
                return
            bitstrg = BitsTrg("varrand")
            for trg in bitstrg:
                trg["end"], trg["ret"] = Forward(), Forward()
                for t in range(27, -1, -1):
                    trg[t] << bt.RawTrigger(
                        conditions=_index.AtLeastX(1, 2**t),
                        actions=bt.SetMemory(trg["ret"] + 16, bt.Add, 18 * (2**t)),
                    )
                trg["end"] << bt.RawTrigger(
                    nextptr=0,
                    actions=[
                        bt.SetMemoryX(trg["ret"], bt.Add, -1, 0x55555555),
                        bt.SetMemoryX(trg["ret"], bt.Add, -1, 0xAAAAAAAA),
                        trg["ret"] << bt.SetMemoryXEPD(0, bt.SetTo, 0, 0),
                    ],
                )

            bits = max((size - 1).bit_length() - 1, 0)
            nptr = Forward()
            if IsEUDVariable(self._epd) and IsEUDVariable(val):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(bitstrg["ret"]) + 4),
                        bt.SetNextPtr(self._epd.GetVTable(), val.GetVTable()),
                        *val.QueueAssignTo(EPD(bitstrg["ret"])),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            elif IsEUDVariable(self._epd):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(bitstrg["ret"]) + 4),
                        bt.SetNextPtr(self._epd.GetVTable(), i.GetVTable()),
                        bt.SetMemory(bitstrg["ret"], bt.SetTo, val),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            elif IsEUDVariable(val):
                bt.RawTrigger(
                    nextptr=val.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, self._epd + 87),
                        *val.QueueAssignTo(EPD(bitstrg["ret"])),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            else:
                bt.RawTrigger(
                    nextptr=i.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, self._epd + 87),
                        bt.SetMemory(bitstrg["ret"], bt.SetTo, val),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            nptr << bt.NextTrigger()

        def ioritem(self, i, val) -> None:
            _bound_check(i)
            if not IsEUDVariable(i):
                ior(self._epd, 18 * i + 87, val)
                return
            bitstrg = BitsTrg("varror")
            for trg in bitstrg:
                trg["end"], trg["ret"] = Forward(), Forward()
                for t in range(27, -1, -1):
                    trg[t] << bt.RawTrigger(
                        conditions=_index.AtLeastX(1, 2**t),
                        actions=bt.SetMemory(trg["ret"] + 16, bt.Add, 18 * (2**t)),
                    )
                trg["end"] << bt.RawTrigger(
                    nextptr=0,
                    actions=[trg["ret"] << bt.SetMemoryXEPD(0, bt.SetTo, ~0, 0)],
                )

            bits = max((size - 1).bit_length() - 1, 0)
            nptr = Forward()
            if IsEUDVariable(self._epd) and IsEUDVariable(val):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(bitstrg["ret"]) + 4),
                        bt.SetNextPtr(self._epd.GetVTable(), val.GetVTable()),
                        *val.QueueAssignTo(EPD(bitstrg["ret"])),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            elif IsEUDVariable(self._epd):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(bitstrg["ret"]) + 4),
                        bt.SetNextPtr(self._epd.GetVTable(), i.GetVTable()),
                        bt.SetMemory(bitstrg["ret"], bt.SetTo, val),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            elif IsEUDVariable(val):
                bt.RawTrigger(
                    nextptr=val.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, self._epd + 87),
                        *val.QueueAssignTo(EPD(bitstrg["ret"])),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            else:
                bt.RawTrigger(
                    nextptr=i.GetVTable(),
                    actions=[
                        bt.SetMemory(bitstrg["ret"] + 16, bt.SetTo, self._epd + 87),
                        bt.SetMemory(bitstrg["ret"], bt.SetTo, val),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(bitstrg["end"], nptr),
                    ],
                )
            nptr << bt.NextTrigger()

        def ixoritem(self, i, val) -> None:
            _bound_check(i)
            if not IsEUDVariable(i):
                ixor(self._epd, 18 * i + 87, val)
                return
            bitstrg = BitsTrg("varrxor")
            for trg in bitstrg:
                trg["end"], trg["ret"] = Forward(), Forward()
                for t in range(27, -1, -1):
                    trg[t] << bt.RawTrigger(
                        conditions=_index.AtLeastX(1, 2**t),
                        actions=bt.SetMemory(0x6509B0, bt.Add, 18 * (2**t)),
                    )
                trg["end"] << bt.RawTrigger(
                    nextptr=cpcache.GetVTable(),
                    actions=[
                        trg["ret"]
                        << bt.SetDeathsX(
                            13, bt.Add, 0, 0, 0x55555555
                        ),  # CurrentPlayer
                        bt.SetDeathsX(13, bt.Add, 0, 0, 0xAAAAAAAA),  # CurrentPlayer
                        cpcache.SetDest(EPD(0x6509B0)),
                    ],
                )

            bits = max((size - 1).bit_length() - 1, 0)
            nptr, trg1, trg2 = Forward(), Forward(), Forward()
            if IsEUDVariable(self._epd) and IsEUDVariable(val):
                trg1 << bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(0x6509B0, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(0x6509B0)),
                        bt.SetNextPtr(self._epd.GetVTable(), val.GetVTable()),
                        *val.QueueAssignTo(EPD(bitstrg["ret"]) + 5),
                        bt.SetNextPtr(val.GetVTable(), trg2),
                        bt.SetNextPtr(cpcache.GetVTable(), nptr),
                    ],
                )
                trg2 << bt.RawTrigger(
                    nextptr=val.GetVTable(),
                    actions=[
                        val.SetDest(EPD(bitstrg["ret"]) + 13),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                    ],
                )
            elif IsEUDVariable(self._epd):
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(0x6509B0, bt.SetTo, 87),
                        *self._epd.QueueAddTo(EPD(0x6509B0)),
                        bt.SetNextPtr(self._epd.GetVTable(), i.GetVTable()),
                        bt.SetMemory(bitstrg["ret"] + 20, bt.SetTo, val),
                        bt.SetMemory(bitstrg["ret"] + 52, bt.SetTo, val),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(cpcache.GetVTable(), nptr),
                    ],
                )
            elif IsEUDVariable(val):
                trg1 << bt.RawTrigger(
                    nextptr=val.GetVTable(),
                    actions=[
                        bt.SetMemory(0x6509B0, bt.SetTo, self._epd + 87),
                        *val.QueueAssignTo(EPD(bitstrg["ret"]) + 5),
                        bt.SetNextPtr(val.GetVTable(), trg2),
                        bt.SetNextPtr(cpcache.GetVTable(), nptr),
                    ],
                )
                trg2 << bt.RawTrigger(
                    nextptr=val.GetVTable(),
                    actions=[
                        val.SetDest(EPD(bitstrg["ret"]) + 13),
                        bt.SetNextPtr(val.GetVTable(), i.GetVTable()),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                    ],
                )
            else:
                bt.RawTrigger(
                    nextptr=self._epd.GetVTable(),
                    actions=[
                        bt.SetMemory(0x6509B0, bt.SetTo, self._epd + 87),
                        bt.SetMemory(bitstrg["ret"] + 20, bt.SetTo, val),
                        bt.SetMemory(bitstrg["ret"] + 52, bt.SetTo, val),
                        *i.QueueAssignTo(_index),
                        bt.SetNextPtr(i.GetVTable(), bitstrg[bits]),
                        bt.SetNextPtr(cpcache.GetVTable(), nptr),
                    ],
                )
            nptr << bt.NextTrigger()

        # FIXME: Add operator?
        def iinvertitem(self, i) -> None:
            _bound_check(i)
            self.ixoritem(i, 0xFFFFFFFF)
            return

        def inotitem(self, i) -> NoReturn:
            _bound_check(i)
            raise AttributeError

        # item comparisons
        def eqitem(self, i, val) -> bt.Condition:
            _bound_check(i)
            if not IsEUDVariable(i):
                return bt.MemoryEPD(self._epd + (18 * i + 87), bt.Exactly, val)
            raise AttributeError

        def neitem(self, i, val):
            _bound_check(i)
            if not IsEUDVariable(i):
                from ...ctrlstru import EUDNot

                return EUDNot(
                    bt.MemoryEPD(self._epd + (18 * i + 87), bt.Exactly, val)
                )
            raise AttributeError

        def leitem(self, i, val) -> bt.Condition:
            _bound_check(i)
            if not IsEUDVariable(i):
                return bt.MemoryEPD(self._epd + (18 * i + 87), bt.AtMost, val)
            raise AttributeError

        def geitem(self, i, val) -> bt.Condition:
            _bound_check(i)
            if not IsEUDVariable(i):
                return bt.MemoryEPD(self._epd + (18 * i + 87), bt.AtLeast, val)
            raise AttributeError

        def ltitem(self, i, val):
            _bound_check(i)
            if not IsEUDVariable(i):
                from ...ctrlstru import EUDNot

                return EUDNot(
                    bt.MemoryEPD(self._epd + (18 * i + 87), bt.AtLeast, val)
                )
            raise AttributeError

        def gtitem(self, i, val):
            _bound_check(i)
            if not IsEUDVariable(i):
                from ...ctrlstru import EUDNot

                return EUDNot(
                    bt.MemoryEPD(self._epd + (18 * i + 87), bt.AtMost, val)
                )
            raise AttributeError

    return _VArray



from helper import *


@TestInstance
def test_blockstru():
    # LoopN
    i = EUDVariable(0)
    if EUDLoopN()(5):
        i += 1
    EUDEndLoopN()
    test_assert("EUDLoopN test", i == 5)

    # While
    i, j = EUDCreateVariables(2)
    if EUDWhileNot()(i >= 100):
        i += 1
        j += 3
    EUDEndWhile()
    test_assert("EUDWhile test", j == 300)



#!/usr/bin/python
# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: I001
# fmt: off

import builtins
import keyword
import types

__version__ = "0.79.6"

from .prelude import *
# TODO: remove lines below
from .core import *
from .utils import *
from .trigger import *
from .ctrlstru import *
from .epscript import *
from .memio import *
from .collections import *
from .scdata import (
    CSprite, EPDCUnitMap,
    Player1, Player2, Player3, Player4, Player5, Player6,
    Player7, Player8, Player9, Player10, Player11, Player12,
)
from .eudlib.mathf.div import f_div_euclid, f_div_floor, f_div_towards_zero
from .eudlib.utilf.gametick import f_getgametick
from .eudlib.utilf.listloop import (
    EUDLoopList,
    EUDLoopNewUnit,
    EUDLoopPlayerUnit,
    EUDLoopUnit,
    EUDLoopUnit2,
)
from .eudlib.utilf.pexist import (
    EUDEndPlayerLoop,
    EUDPlayerLoop,
)
from .string import *
from .string.cpprint import f_cpstr_adddw, f_cpstr_addptr, f_cpstr_addstr
from .maprw import *
from .qgc import *
from .trigtrg.runtrigtrg import (
    GetFirstTrigTrigger,
    GetLastTrigTrigger,
    RunTrigTrigger,
    TrigTriggerBegin,
    TrigTriggerEnd,
)

# remove modules from __all__
_old_globals = [
    "keyword",
    "__file__",
    "types",
    "__doc__",
    "__version__",
    "builtins",
    "__cached__",
    "__name__",
    "__loader__",
    "__spec__",
    "__path__",
    "__package__",
    "__builtins__",
]
_alllist = []
for _k, _v in dict(globals()).items():
    if _k in _old_globals:
        continue
    elif _k != "stocktrg" and isinstance(_v, types.ModuleType):
        continue
    elif _k[0] == "_":
        continue
    _alllist.append(_k)

__all__ = _alllist

del _k
del _v


def eudplibVersion():  # noqa: N802
    return __version__


_alllist.append("eudplibVersion")


from .epscript import epscompile

epscompile._set_eps_globals(_alllist)
epscompile._set_py_keywords(keyword.kwlist)
epscompile._set_py_builtins(dir(builtins))



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ... import utils as ut
from ...localize import _
from .. import allocator as ac
from .. import rawtrigger as rt
from ..eudstruct import EUDStruct
from ..eudstruct.vararray import _InternalVArray
from ..variable import EUDVariable, SetVariables, VProc
from .eudfuncn import EUDFuncN
from .eudtypedfuncn import _apply_types

#
# Common argument / returns storage
#


def _get_arg_storage(argn, _argstorage_dict={}):
    """Get common arguments storage for argn"""
    if argn not in _argstorage_dict:
        _argstorage_dict[argn] = [EUDVariable() for _ in range(argn)]
    return _argstorage_dict[argn]


def _get_ret_storage(retn, _retstorage_dict={}):
    """Get common returns storage for retn"""
    if retn not in _retstorage_dict:
        _retstorage_dict[retn] = [EUDVariable() for _ in range(retn)]
    return _retstorage_dict[retn]


def _fill_arguments(f):
    """Copy values from common argument storage to f._args"""
    if f._argn:
        arg_storage = _get_arg_storage(f._argn)
        for farg, arg in zip(f._fargs, arg_storage):
            VProc(arg, arg.QueueAssignTo(farg))


def _fill_returns(f):
    """Copy values from f_rets to common returns storage"""
    if f._retn:
        ret_storage = _get_ret_storage(f._retn)
        for fret, ret in zip(f._frets, ret_storage):
            VProc(fret, fret.QueueAssignTo(ret))


def _call_func_body(fstart, fend):
    """Call function's body triggers"""
    fcallend = ac.Forward()

    rt.RawTrigger(nextptr=fstart, actions=[rt.SetNextPtr(fend, fcallend)])

    fcallend << rt.NextTrigger()


def _create_indirect_caller(f, _caller_dict={}):
    """Create function caller using common argument/return storage"""

    # Cache function in _caller_dict. If uncached,
    if f not in _caller_dict:
        rt.PushTriggerScope()
        caller_start = rt.NextTrigger()
        _fill_arguments(f)
        _call_func_body(f._fstart, f._fend)
        _fill_returns(f)
        caller_end = rt.RawTrigger()
        rt.PopTriggerScope()

        _caller_dict[f] = (caller_start, caller_end)

    return _caller_dict[f]


# ---------------------------------


def EUDTypedFuncPtr(argtypes, rettypes):  # noqa: N802
    argn = len(argtypes)
    retn = len(rettypes)

    class PtrDataClass(EUDStruct):
        _fields_ = ("_fstart", "_fendnext_epd")

        def __init__(self, _from=None) -> None:
            if _from is not None and isinstance(_from, EUDFuncN):
                # Statically generate with EUDFuncN
                self._check_valid_function(_from)
                f_idcstart, f_idcend = _create_indirect_caller(_from)
                super().__init__(
                    _from=_InternalVArray(2)([f_idcstart, ut.EPD(f_idcend + 4)])
                )
            else:
                # Cast from EUDVariable or ConstExpr
                super().__init__(_from=_from)

        @classmethod
        def cast(cls, _from):
            return cls(_from=_from)

        def _check_valid_function(self, f):
            ut.ep_assert(
                isinstance(f, EUDFuncN), _("{} is not an EUDFuncN").format(f)
            )
            if not f._fstart:
                f._create_func_body()

            f_argn = f._argn
            f_retn = f._retn
            ut.ep_assert(
                argn == f_argn,
                _("Function requires {} arguments (Expected {})").format(
                    f_argn, argn
                ),
            )
            ut.ep_assert(
                retn == f_retn,
                _("Function returns {} values (Expected {})").format(f_retn, retn),
            )

        def setFunc(self, f):  # noqa: N802
            """Set function pointer's target to function

            :param f: Target function
            """
            try:
                self._fstart, self._fendnext_epd = f._fstart, f._fendnext_epd

            except AttributeError:
                self._check_valid_function(f)

                # Build actions
                f_idcstart, f_idcend = _create_indirect_caller(f)
                self._fstart = f_idcstart
                self._fendnext_epd = ut.EPD(f_idcend + 4)

        def __lshift__(self, rhs):
            self.setFunc(rhs)

        def __call__(self, *args):
            """Call target function with given arguments"""

            args = _apply_types(argtypes, args)

            if argn:
                arg_storage = _get_arg_storage(argn)
                SetVariables(arg_storage, args)

            # Call function
            t, a = ac.Forward(), ac.Forward()
            SetVariables(
                [ut.EPD(t + 4), ut.EPD(a + 16)],
                [self._fstart, self._fendnext_epd],
            )

            fcallend = ac.Forward()
            t << rt.RawTrigger(actions=[a << rt.SetNextPtr(0, fcallend)])
            fcallend << rt.NextTrigger()

            if retn:
                tmp_rets = [EUDVariable() for _ in range(retn)]
                ret_storage = _get_ret_storage(retn)
                SetVariables(tmp_rets, ret_storage)
                tmp_rets = _apply_types(rettypes, tmp_rets)
                return ut.List2Assignable(tmp_rets)

    return PtrDataClass


def EUDFuncPtr(argn, retn):  # noqa: N802
    return EUDTypedFuncPtr([None] * argn, [None] * retn)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import types
from typing import Any

from ... import core as c
from ... import ctrlstru as cs
from ... import eudlib as sf
from ... import trigger as t
from ... import utils as ut
from ...core import RawTrigger
from .btinliner import t_start_end

_inline_globals = {}


def compute_base_inline_code_globals() -> None:
    """
    Return list of globals inline eudplib code can call.
    """

    global _inline_globals

    # Add eudplib functions
    g = {}
    modules = [c, cs, sf, ut, t]
    for module in modules:
        for k, v in module.__dict__.items():
            if isinstance(v, types.ModuleType):
                continue
            if k[0] == "_":
                continue

            g[k] = v

    _inline_globals = g


def get_inline_code_globals() -> dict[str, Any]:
    """
    Return list of globals inline eudplib code can call.
    """

    g = _inline_globals.copy()

    # Add custom registered functions
    for k, v in c.GetEUDNamespace().items():
        g[k] = v

    return g


def _compile_inline_code(code: str) -> t_start_end:
    _code = compile(code, "<string>", "exec")

    if c.PushTriggerScope():
        tstart = RawTrigger(actions=c.SetDeaths(0, c.SetTo, 0, 0))
        exec(_code, get_inline_code_globals(), {})
        tend = RawTrigger()
    c.PopTriggerScope()

    return tstart, tend



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .eudarray import EUDArray


def EUDStack(basetype=None):  # noqa: N802
    class _EUDStack(c.EUDStruct):
        _fields_ = (("data", EUDArray), "pos")

        def constructor(self, size):
            self.data = EUDArray(size)
            self.pos = 0

        def push(self, value):
            self.data[self.pos] = value
            self.pos += 1

        def pop(self):
            self.pos -= 1
            data = self.data[self.pos]
            if basetype:
                data = basetype.cast(data)
            return data

        def empty(self):
            return self.pos == 0

    return _EUDStack



## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) var x: CUnit = 0;
x = _TYGV([CUnit], lambda: [0])
# (Line 2) function test_cunit1() {
@EUDFunc
def f_test_cunit1():
    # (Line 3) x = 0;
    x << (0)
    # (Line 4) const get_epd = py_eval("lambda x: x if x._value._is_epd() else EPD(x)");
    get_epd = eval("lambda x: x if x._value._is_epd() else EPD(x)")
    # (Line 5) const cspritePtr = 0x629D98 + 36 * 1;  // Sprite Table
    cspritePtr = 0x629D98 + 36 * 1
    # (Line 6) const sp = [0x2468ACE0, 0x13579BDF, 0xFEDCBA98, 0x76543210];
    sp = _ARR(FlattenList([0x2468ACE0, 0x13579BDF, 0xFEDCBA98, 0x76543210]))
    # (Line 7) repmovsd_epd(EPD(cspritePtr), get_epd(sp), 4);
    f_repmovsd_epd(EPD(cspritePtr), get_epd(sp), 4)
    # (Line 8) const a = [
    # (Line 9) 0,  -1,  -2,  cspritePtr,  -4,  -5,  -6,  -7,  -8, -9,
    # (Line 10) -10, -11, -12, -13, -14, -15, -16, -17, -18, 0x12345607,
    # (Line 11) 0x89ABCDEF, 0xEDACEDAC, 0xEDACEDAC, 0xEDACEDAC, 0x220011, 0x440033, 0xFFFF0055, 0, 0, 0,
    # (Line 12) 3, 3, 3, 3, 3, 3, 3, 3, 0x220011, 0x440033,
    # (Line 13) 0xFFFF0055, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    # (Line 14) 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    # (Line 15) ];
    a = _ARR(FlattenList([0, -1, -2, cspritePtr, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -14, -15, -16, -17, -18, 0x12345607, 0x89ABCDEF, 0xEDACEDAC, 0xEDACEDAC, 0xEDACEDAC, 0x220011, 0x440033, 0xFFFF0055, 0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3, 0x220011, 0x440033, 0xFFFF0055, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]))
    # (Line 16) const u = CUnit.cast(EUDVariable(get_epd(a)));
    u = CUnit.cast(EUDVariable(get_epd(a)))
    # (Line 17) py_exec("\n\
    # (Line 21) ");
    exec("\nfrom helper import expect_error\nwith expect_error(AttributeError):\n    u.zzt = 1\n")
    # (Line 22) var ret = 0;
    ret = _TYLV([None], [0])
    # (Line 23) if (u.order == 0x56) ret += 1;
    if EUDIf()(_ATTC(u, 'order') == 0x56):
        ret.__iadd__(1)
        # (Line 24) if (u.owner == P8) ret += 2;
    EUDEndIf()
    if EUDIf()(_ATTC(u, 'owner') == P8):
        ret.__iadd__(2)
        # (Line 25) if (u.orderState == 0x34) ret += 4;
    EUDEndIf()
    if EUDIf()(_ATTC(u, 'orderState') == 0x34):
        ret.__iadd__(4)
        # (Line 26) if (u.orderSignal == 0x12) ret += 8;
    EUDEndIf()
    if EUDIf()(_ATTC(u, 'orderSignal') == 0x12):
        ret.__iadd__(8)
        # (Line 28) if (u.orderUnitType == 0xCDEF) ret += 16;
    EUDEndIf()
    if EUDIf()(_ATTC(u, 'orderUnitType') == 0xCDEF):
        ret.__iadd__(16)
        # (Line 29) wwrite_epd(get_epd(a) + 0x50/4, 0, $U("Artanis"));
    EUDEndIf()
    f_wwrite_epd(get_epd(a) + 0x50 // 4, 0, EncodeUnit("Artanis"))
    # (Line 30) if (u.orderUnitType == py_str("Artanis")) ret += 32;
    if EUDIf()(_ATTC(u, 'orderUnitType') == str("Artanis")):
        ret.__iadd__(32)
        # (Line 31) u.set_color(P12);
    EUDEndIf()
    u.set_color(P12)
    # (Line 32) if (Memory(cspritePtr + 8, Exactly, 0xFE0BBA98)) ret += 64;
    if EUDIf()(Memory(cspritePtr + 8, Exactly, 0xFE0BBA98)):
        ret.__iadd__(64)
        # (Line 33) u.cgive(P2);
    EUDEndIf()
    u.cgive(P2)
    # (Line 34) if (u.owner == P2) ret += 128;
    if EUDIf()(_ATTC(u, 'owner') == P2):
        ret.__iadd__(128)
        # (Line 36) u.die();
    EUDEndIf()
    u.die()
    # (Line 37) if (a[0x4C/4] == 0x12340001) ret += 256;
    if EUDIf()(_ARRC(a, 0x4C // 4) == 0x12340001):
        ret.__iadd__(256)
        # (Line 38) if (u.is_dying()) ret += 512;
    EUDEndIf()
    if EUDIf()(u.is_dying()):
        ret.__iadd__(512)
        # (Line 39) if (!u.are_buildq_empty()) ret += 1024;
    EUDEndIf()
    if EUDIf()(u.are_buildq_empty(), neg=True):
        ret.__iadd__(1024)
        # (Line 40) if (u.check_buildq(EUDVariable(0x11))) ret += 2048;
    EUDEndIf()
    if EUDIf()(u.check_buildq(EUDVariable(0x11))):
        ret.__iadd__(2048)
        # (Line 42) if (u.check_buildq("Terran Medic")) ret += 4096;
    EUDEndIf()
    if EUDIf()(u.check_buildq("Terran Medic")):
        ret.__iadd__(4096)
        # (Line 43) if (u.check_buildq(EUDVariable(0x33))) ret += 8192;
    EUDEndIf()
    if EUDIf()(u.check_buildq(EUDVariable(0x33))):
        ret.__iadd__(8192)
        # (Line 44) if (u.check_buildq("Protoss Archon")) ret += 16384;
    EUDEndIf()
    if EUDIf()(u.check_buildq("Protoss Archon")):
        ret.__iadd__(16384)
        # (Line 45) if (u.check_buildq("Protoss Scarab")) ret += 32768;
    EUDEndIf()
    if EUDIf()(u.check_buildq("Protoss Scarab")):
        ret.__iadd__(32768)
        # (Line 47) u.reset_buildq();
    EUDEndIf()
    u.reset_buildq()
    # (Line 48) if (u.are_buildq_empty()) ret += 65536;
    if EUDIf()(u.are_buildq_empty()):
        ret.__iadd__(65536)
        # (Line 49) var d = u.movementFlags;
    EUDEndIf()
    d = _TYLV([None], [u.movementFlags])
    # (Line 50) if (d == 0xF8) ret += 1 << 17;
    if EUDIf()(d == 0xF8):
        ret.__iadd__(_LSH(1,17))
        # (Line 51) u.movementFlags += 1;
    EUDEndIf()
    _ATTW(u, 'movementFlags').__iadd__(1)
    # (Line 52) if (u.movementFlags == 0xF9) ret += 1 << 18;
    if EUDIf()(_ATTC(u, 'movementFlags') == 0xF9):
        ret.__iadd__(_LSH(1,18))
        # (Line 53) return ret;
    EUDEndIf()
    EUDReturn(ret)
    # (Line 54) }
    # (Line 55) function test_cunit2() {

@EUDFunc
def f_test_cunit2():
    # (Line 56) const u = CUnit(EUDVariable(EPD(0x59CCA8 + 336)));
    u = CUnit(EUDVariable(EPD(0x59CCA8 + 336)))
    # (Line 57) var ret = 0;
    ret = _TYLV([None], [0])
    # (Line 59) u.orderQueueCount = 123;
    _ATTW(u, 'orderQueueCount') << (123)
    # (Line 60) if (u.orderQueueCount == 123) ret += 1;
    if EUDIf()(_ATTC(u, 'orderQueueCount') == 123):
        ret.__iadd__(1)
        # (Line 61) u.orderQueueCount = 0;
    EUDEndIf()
    _ATTW(u, 'orderQueueCount') << (0)
    # (Line 62) if (u.orderQueueCount == 0) ret += 2;
    if EUDIf()(_ATTC(u, 'orderQueueCount') == 0):
        ret.__iadd__(2)
        # (Line 64) u.currentDirection1 = 234;
    EUDEndIf()
    _ATTW(u, 'currentDirection1') << (234)
    # (Line 65) if (u.currentDirection1 == 234) ret += 4;
    if EUDIf()(_ATTC(u, 'currentDirection1') == 234):
        ret.__iadd__(4)
        # (Line 66) u.currentDirection1 = 0;
    EUDEndIf()
    _ATTW(u, 'currentDirection1') << (0)
    # (Line 67) if (u.currentDirection1 == 0) ret += 8;
    if EUDIf()(_ATTC(u, 'currentDirection1') == 0):
        ret.__iadd__(8)
        # (Line 69) u.turnRadius = 34;
    EUDEndIf()
    _ATTW(u, 'turnRadius') << (34)
    # (Line 70) if (u.turnRadius == 34) ret += 16;
    if EUDIf()(_ATTC(u, 'turnRadius') == 34):
        ret.__iadd__(16)
        # (Line 71) u.turnRadius = 0;
    EUDEndIf()
    _ATTW(u, 'turnRadius') << (0)
    # (Line 72) if (u.turnRadius == 0) ret += 32;
    if EUDIf()(_ATTC(u, 'turnRadius') == 0):
        ret.__iadd__(32)
        # (Line 74) u.velocityDirection1 = 45;
    EUDEndIf()
    _ATTW(u, 'velocityDirection1') << (45)
    # (Line 75) if (u.velocityDirection1 == 45) ret += 64;
    if EUDIf()(_ATTC(u, 'velocityDirection1') == 45):
        ret.__iadd__(64)
        # (Line 76) u.velocityDirection1 = 0;
    EUDEndIf()
    _ATTW(u, 'velocityDirection1') << (0)
    # (Line 77) if (u.velocityDirection1 == 0) ret += 128;
    if EUDIf()(_ATTC(u, 'velocityDirection1') == 0):
        ret.__iadd__(128)
        # (Line 79) u.energy = 567;
    EUDEndIf()
    _ATTW(u, 'energy') << (567)
    # (Line 80) if (u.energy == 567) ret += 256;
    if EUDIf()(_ATTC(u, 'energy') == 567):
        ret.__iadd__(256)
        # (Line 81) u.energy = 0;
    EUDEndIf()
    _ATTW(u, 'energy') << (0)
    # (Line 82) if (u.energy == 0) ret += 512;
    if EUDIf()(_ATTC(u, 'energy') == 0):
        ret.__iadd__(512)
        # (Line 84) u.acceleration = 678;
    EUDEndIf()
    _ATTW(u, 'acceleration') << (678)
    # (Line 85) if (u.acceleration == 678) ret += 1024;
    if EUDIf()(_ATTC(u, 'acceleration') == 678):
        ret.__iadd__(1024)
        # (Line 86) u.acceleration = 0;
    EUDEndIf()
    _ATTW(u, 'acceleration') << (0)
    # (Line 87) if (u.acceleration == 0) ret += 2048;
    if EUDIf()(_ATTC(u, 'acceleration') == 0):
        ret.__iadd__(2048)
        # (Line 89) u.remove();
    EUDEndIf()
    u.remove()
    # (Line 90) if (u.order == 0 && u.userActionFlags & 4) ret += 4096;
    if EUDIf()(EUDSCAnd()(_ATTC(u, 'order') == 0)(u.userActionFlags & 4)()):
        ret.__iadd__(4096)
        # (Line 92) return ret;
    EUDEndIf()
    EUDReturn(ret)
    # (Line 93) }



from helper import *


@TestInstance
def test_cpmemio():
    a = Db(i2b4(0x1234) + i2b4(0x5678) + i2b4(0x9ABC))

    f_setcurpl(EPD(a))

    # Normal reading
    ptr, epd = f_dwepdread_cp(0)
    test_equality(
        "f_dwread_cp with cpo=0", [ptr, 0x58A364 + 4 * epd], [0x1234, 0x1234]
    )

    a, b, c = (f_dwread_cp(0), f_dwread_cp(1), f_dwread_cp(2))

    test_equality("f_dwread_cp", [a, b, c], [0x1234, 0x5678, 0x9ABC])

    # Issue #4: This should at least compile.
    v = EUDVariable(0)
    f_dwepdread_cp(v)

    # Writing
    f_dwwrite_cp(1, 0x1111)
    a, b, c = (f_dwread_cp(0), f_dwread_cp(1), f_dwread_cp(2))

    test_equality("f_dwread_cp", [a, b, c], [0x1234, 0x1111, 0x9ABC])

    # byte/word reading
    f_dwadd_cp(0, 0xABCD0000)
    a, b, c = (f_wread_cp(0, 3), f_wread_cp(1, 1), f_bread_cp(2, 1))
    test_equality("f_wread_cp & f_bread_cp", [a, b, c], [0x11AB, 0x11, 0x9A])

    # byte/word writing
    f_wwrite_cp(0, 1, 0x4567)
    f_wwrite_cp(1, 2, 0x8765)
    f_bwrite_cp(2, 1, 0x46)
    a, b, c = (f_wread_cp(0, 2), f_bread_cp(1, 3), f_bread_cp(2, 1))
    test_equality("f_bwrite_cp & f_wwrite_cp", [a, b, c], [0xAB45, 0x87, 0x46])

    f_setcurpl(P1)



from typing import Literal, TypeAlias

DefTechDict: dict[str, int] = {
    "Stim Packs": 0,
    "Lockdown": 1,
    "EMP Shockwave": 2,
    "Spider Mines": 3,
    "Scanner Sweep": 4,
    "Tank Siege Mode": 5,
    "Defensive Matrix": 6,
    "Irradiate": 7,
    "Yamato Gun": 8,
    "Cloaking Field": 9,
    "Personnel Cloaking": 10,
    "Burrowing": 11,
    "Infestation": 12,
    "Spawn Broodlings": 13,
    "Dark Swarm": 14,
    "Plague": 15,
    "Consume": 16,
    "Ensnare": 17,
    "Parasite": 18,
    "Psionic Storm": 19,
    "Hallucination": 20,
    "Recall": 21,
    "Stasis Field": 22,
    "Archon Warp": 23,
    "Restoration": 24,
    "Disruption Web": 25,
    "Unused26": 26,
    "Mind Control": 27,
    "Dark Archon Meld": 28,
    "Feedback": 29,
    "Optical Flare": 30,
    "Maelstrom": 31,
    "Lurker Aspect": 32,
    "Unused33": 33,
    "Healing": 34,
    "Unused35": 35,
    "Unused36": 36,
    "Unused37": 37,
    "Unused38": 38,
    "Unused39": 39,
    "Unused40": 40,
    "Unused41": 41,
    "Unused42": 42,
    "Unused43": 43,
}

DefaultTech: TypeAlias = Literal[
    "Stim Packs",
    "Lockdown",
    "EMP Shockwave",
    "Spider Mines",
    "Scanner Sweep",
    "Tank Siege Mode",
    "Defensive Matrix",
    "Irradiate",
    "Yamato Gun",
    "Cloaking Field",
    "Personnel Cloaking",
    "Burrowing",
    "Infestation",
    "Spawn Broodlings",
    "Dark Swarm",
    "Plague",
    "Consume",
    "Ensnare",
    "Parasite",
    "Psionic Storm",
    "Hallucination",
    "Recall",
    "Stasis Field",
    "Archon Warp",
    "Restoration",
    "Disruption Web",
    "Unused26",
    "Mind Control",
    "Dark Archon Meld",
    "Feedback",
    "Optical Flare",
    "Maelstrom",
    "Lurker Aspect",
    "Unused33",
    "Healing",
    "Unused35",
    "Unused36",
    "Unused37",
    "Unused38",
    "Unused39",
    "Unused40",
    "Unused41",
    "Unused42",
    "Unused43",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from random import random

from ... import utils as ut
from ...core.mapdata.chktok import CHK
from ...core.rawtrigger import SetTo
from ...scdata import AllPlayers
from ...trigtrg import trigtrg as tt
from .btinliner import (
    get_executing_players,
    get_trigger_size,
    inline_codify_binary_trigger,
    inline_codify_binary_triggers,
    t_start_end,
    try_share_trigger,
)
from .ilccompile import _compile_inline_code, compute_base_inline_code_globals

_inline_codes: list[tuple[int, t_start_end]] = []
_inlining_rate: float = 1.0
_cutoff_rate: list[float] = [1 + (i - 2) / 3 for i in range(9)]


def PRT_SetInliningRate(rate: float) -> None:  # noqa: N802
    """Set how much triggers will be inlined into STR section."""
    global _inlining_rate
    _inlining_rate = rate


def _preprocess_inline_code(chkt: CHK) -> None:
    global _inline_codes
    trig_section = chkt.getsection("TRIG")
    _inline_codes, trig_section = _preprocess_trig_section(trig_section)
    chkt.setsection("TRIG", trig_section)


def _preprocess_trig_section(
    trig_section: bytes,
) -> tuple[list[tuple[int, t_start_end]], bytes]:
    """Fetch inline codes & compiles them"""
    compute_base_inline_code_globals()
    if _inlining_rate >= 1.0:
        return _consecutive_inline_trig_section(trig_section)

    inline_codes: list[tuple[int, t_start_end]] = []
    trig_segments: list[bytes] = []
    for i in range(0, len(trig_section), 2400):
        trig_segment = trig_section[i : i + 2400]
        if len(trig_segment) != 2400:
            continue

        propv = ut.b2i4(trig_segment, 320 + 2048)

        decoded = _dispatch_inline_code(inline_codes, trig_segment)
        if decoded:
            trig_segment = decoded

        elif propv < 0x80000000 and random() < _inlining_rate:
            trig_segment = _inlinify_normal_trigger(inline_codes, trig_segment)

        trig_segments.append(trig_segment)

    """
    This is rather hard to explain, but we need blank trigger.

    If inline_eudplib trigger is at the last of the trigger, then its
    nextptr should be modified to codeStart.

    But after flipprop works, RunTrigTrigger engine will re-change its nextptr
    to somewhere else, where everything quits.

    So we need 'normal' trigger at the last of TRIG triggers for every player.
    """
    trig_segments.append(tt.Trigger(players=[AllPlayers]))

    trig_section = b"".join(trig_segments)
    return inline_codes, trig_section


def _consecutive_inline_trig_section(
    trig_section: bytes,
) -> tuple[list[tuple[int, t_start_end]], bytes]:
    inline_codes: list[tuple[int, t_start_end]] = []
    trig_segments: list[bytes] = []
    ptriggers: list[list[bytes]] = [[] for _ in range(8)]

    def append_ptriggers(p):
        if ptriggers[p]:
            func = inline_codify_binary_triggers(ptriggers[p])
            trig_segment = _create_inline_code_dispatcher(inline_codes, func, 1 << p)
            trig_segments.append(trig_segment)
            ptriggers[p].clear()

    for i in range(0, len(trig_section), 2400):
        trig_segment = trig_section[i : i + 2400]
        if len(trig_segment) != 2400:
            continue

        propv = ut.b2i4(trig_segment, 320 + 2048)
        executing_players = get_executing_players(trig_segment)

        decoded = _dispatch_inline_code(inline_codes, trig_segment)
        if decoded:
            trig_segment = decoded

        elif propv < 0x80000000:
            player_count = executing_players.count(True)
            if player_count >= 2:
                index_or_trig = try_share_trigger(trig_segment)
                if isinstance(index_or_trig, bytes):
                    size = get_trigger_size(index_or_trig)
                    if size * player_count * _cutoff_rate[player_count] > 2400:
                        for p in range(8):
                            if executing_players[p]:
                                append_ptriggers(p)
                        trig_segments.append(index_or_trig)
                        continue
            for p in range(8):
                if executing_players[p]:
                    ptriggers[p].append(trig_segment)
            continue
        for p in range(8):
            if executing_players[p]:
                append_ptriggers(p)
        trig_segments.append(trig_segment)
    for p in range(8):
        append_ptriggers(p)
    trig_segments.append(tt.Trigger(players=[AllPlayers]))

    trig_section = b"".join(trig_segments)
    return inline_codes, trig_section


def _get_inline_code_list() -> list[tuple[int, t_start_end]]:
    """Get list of compiled inline_eudplib code"""
    return _inline_codes


def GetInlineCodePlayerList(btrigger: bytes) -> int | None:  # noqa: N802
    # Check if effplayer & current_action is empty
    for player in range(28):
        if btrigger[320 + 2048 + 4 + player] != 0:
            return None

    # trg.cond[0].condtype != 0
    if btrigger[15] != 0:
        return None
    # trg.act[0].acttype != 0
    if btrigger[346] != 0:
        return None

    return ut.b2i4(btrigger, 24)


def _dispatch_inline_code(
    inline_codes: list[tuple[int, t_start_end]], trigger_bytes: bytes
) -> bytearray | None:
    """Check if trigger segment has special data."""
    magic_code = ut.b2i4(trigger_bytes, 20)
    if magic_code != 0x10978D4A:
        return None

    player_code = GetInlineCodePlayerList(trigger_bytes)
    if not player_code:
        return None

    data = trigger_bytes[20:320] + trigger_bytes[352:2372]
    code_data = ut.b2u(data[8:]).rstrip("\0")

    # Compile code
    func = _compile_inline_code(code_data)
    return _create_inline_code_dispatcher(inline_codes, func, player_code)


def _inlinify_normal_trigger(
    inline_codes: list[tuple[int, t_start_end]], trigger_bytes: bytes
) -> bytearray:
    """Inlinify normal binary triggers"""
    player_code = 0
    for i in range(27):
        if trigger_bytes[320 + 2048 + 4 + i]:
            player_code |= 1 << i

    func = inline_codify_binary_trigger(trigger_bytes)
    return _create_inline_code_dispatcher(inline_codes, func, player_code)


def _create_inline_code_dispatcher(
    inline_codes: list[tuple[int, t_start_end]],
    func: t_start_end,
    player_code: int,
) -> bytearray:
    """Create link from TRIG list to STR trigger."""
    func_id = len(inline_codes) + 1024
    inline_codes.append((func_id, func))

    # Return new trigger
    new_trigger = bytearray(2400)

    # Apply effplayer
    for player in range(27):
        if player_code & (1 << player):
            new_trigger[320 + 2048 + 4 + player] = 1

    # Apply 4 SetDeaths
    setdeaths_template = tt.SetDeaths(0, SetTo, 0, 0)
    new_trigger[320 + 32 * 0 : 320 + 32 * 1] = setdeaths_template
    new_trigger[320 + 32 * 1 : 320 + 32 * 2] = setdeaths_template
    new_trigger[320 + 32 * 2 : 320 + 32 * 3] = setdeaths_template
    new_trigger[320 + 32 * 3 : 320 + 32 * 4] = setdeaths_template

    # Apply flag
    new_trigger[0:4] = ut.i2b4(func_id)
    new_trigger[2368:2372] = b"\0\0\0\x10"

    return new_trigger



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing import TypeAlias

from ...bindings._rust import allocator as al
from ...utils import ExprProxy
from .rlocint import RlocInt_C

ConstExpr = al.ConstExpr
Forward = al.Forward
Evaluable: TypeAlias = ConstExpr | int | ExprProxy[ConstExpr] | RlocInt_C
Evaluate = al.Evaluate
IsConstExpr = al.IsConstExpr



## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) object Test {
# (Line 2) var x, y, z;
class Test(EUDStruct):
    # (Line 3) function setX(x) { this.x = x; }
    @EUDMethod
    def setX(this, x):
        _ATTW(this, 'x') << (x)
        # (Line 4) function add(other: selftype) {

    @EUDTypedMethod([selftype])
    def add(this, other):
        # (Line 5) this.x += other.x;
        _ATTW(this, 'x').__iadd__(other.x)
        # (Line 6) this.y += other.y;
        _ATTW(this, 'y').__iadd__(other.y)
        # (Line 7) this.z += other.z;
        _ATTW(this, 'z').__iadd__(other.z)
        # (Line 8) }
        # (Line 9) };

    # (Line 10) function test_object() {
    _fields_ = [
        'x',
        'y',
        'z',
    ]

@EUDFunc
def f_test_object():
    # (Line 11) const a = Test();
    a = Test()
    # (Line 12) a.x, a.y = 5, 7;
    _SV([_ATTW(a, 'x'), _ATTW(a, 'y')], [5, 7])
    # (Line 14) const b = Test.cast(a);
    b = Test.cast(a)
    # (Line 15) const c = Test.cast(b);
    c = Test.cast(b)
    # (Line 17) b.z = 8;
    _ATTW(b, 'z') << (8)
    # (Line 18) c.x = 3;
    _ATTW(c, 'x') << (3)
    # (Line 21) var ret = a ? 1 : 0;
    ret = _TYLV([None], [EUDTernary(a)(1)(0)])
    # (Line 22) if (a.x == b.x && b.x == c.x) ret += 2;
    if EUDIf()(EUDSCAnd()(_ATTC(a, 'x') == b.x)(_ATTC(b, 'x') == c.x)()):
        ret.__iadd__(2)
        # (Line 23) if (a.y == b.y && b.y == c.y) ret += 4;
    EUDEndIf()
    if EUDIf()(EUDSCAnd()(_ATTC(a, 'y') == b.y)(_ATTC(b, 'y') == c.y)()):
        ret.__iadd__(4)
        # (Line 24) if (a.z == b.z && b.z == c.z) ret += 8;
    EUDEndIf()
    if EUDIf()(EUDSCAnd()(_ATTC(a, 'z') == b.z)(_ATTC(b, 'z') == c.z)()):
        ret.__iadd__(8)
        # (Line 25) if (a.x == 3 && b.y == 7 && c.z == 8) ret += 16;
    EUDEndIf()
    if EUDIf()(EUDSCAnd()(_ATTC(a, 'x') == 3)(_ATTC(b, 'y') == 7)(_ATTC(c, 'z') == 8)()):
        ret.__iadd__(16)
        # (Line 27) const d = a.copy();
    EUDEndIf()
    d = a.copy()
    # (Line 28) a.add(d);
    a.add(d)
    # (Line 29) if (c.x == 6 && a.y == 14 && b.z == 16) ret += 32;
    if EUDIf()(EUDSCAnd()(_ATTC(c, 'x') == 6)(_ATTC(a, 'y') == 14)(_ATTC(b, 'z') == 16)()):
        ret.__iadd__(32)
        # (Line 30) if (d.x == 3 && d.y == 7 && d.z == 8) ret += 64;
    EUDEndIf()
    if EUDIf()(EUDSCAnd()(_ATTC(d, 'x') == 3)(_ATTC(d, 'y') == 7)(_ATTC(d, 'z') == 8)()):
        ret.__iadd__(64)
        # (Line 32) var e = a;
    EUDEndIf()
    e = _TYLV([None], [a])
    # (Line 33) Test.cast(e).setX(1);
    Test.cast(e).setX(1)
    # (Line 34) if (a.x == 1) ret += 128;
    if EUDIf()(_ATTC(a, 'x') == 1):
        ret.__iadd__(128)
        # (Line 35) Test.cast(e).add(d);
    EUDEndIf()
    Test.cast(e).add(d)
    # (Line 36) if (b.x == 4 && c.y == 21 && a.z == 24) ret += 256;
    if EUDIf()(EUDSCAnd()(_ATTC(b, 'x') == 4)(_ATTC(c, 'y') == 21)(_ATTC(a, 'z') == 24)()):
        ret.__iadd__(256)
        # (Line 38) return ret;  // 511
    EUDEndIf()
    EUDReturn(ret)
    # (Line 39) }
    # (Line 41) object Triangle;

# (Line 42) object Coord { var x, y; };
class Coord(EUDStruct):
    # (Line 43) object Triangle {
    _fields_ = [
        'x',
        'y',
    ]

# (Line 44) var p: Coord * 5;
class Triangle(EUDStruct):
    # (Line 45) var q;
    # (Line 46) };
    # (Line 48) function test_nested_object() {
    _fields_ = [
        ('p', Coord * 5),
        'q',
    ]

@EUDFunc
def f_test_nested_object():
    # (Line 49) const a = Triangle();
    a = Triangle()
    # (Line 50) a.p = (Coord * 5)();
    _ATTW(a, 'p') << ((Coord * 5)())
    # (Line 51) foreach(i : py_range(5)) {
    for i in range(5):
        # (Line 52) a.p[i] = Coord();
        _ARRW(a.p, i) << (Coord())
        # (Line 53) }
        # (Line 55) a.p[0].x = 1;

    _ATTW(a.p[0], 'x') << (1)
    # (Line 56) a.p[1].y = 2;
    _ATTW(a.p[1], 'y') << (2)
    # (Line 57) a.p[1].y += 2;
    _ATTW(a.p[1], 'y').__iadd__(2)
    # (Line 58) a.q = 3;
    _ATTW(a, 'q') << (3)
    # (Line 60) var ret = list(a.p[0].x == 1, a.p[0].y == 0, a.p[1].y == 4, a.q == 3) ? 1 : 0;
    ret = _TYLV([None], [EUDTernary(FlattenList([_ATTC(a.p[0], 'x') == 1, _ATTC(a.p[0], 'y') == 0, _ATTC(a.p[1], 'y') == 4, _ATTC(a, 'q') == 3]))(1)(0)])
    # (Line 62) const b = a.copy();
    b = a.copy()
    # (Line 63) b.p[0].x = 5;
    _ATTW(b.p[0], 'x') << (5)
    # (Line 64) b.q = 2;
    _ATTW(b, 'q') << (2)
    # (Line 65) if (a.p[0].x == 5) ret += 2;
    if EUDIf()(_ATTC(a.p[0], 'x') == 5):
        ret.__iadd__(2)
        # (Line 66) if (b.p[0].x == 5) ret += 4;
    EUDEndIf()
    if EUDIf()(_ATTC(b.p[0], 'x') == 5):
        ret.__iadd__(4)
        # (Line 67) if (b.p[0].y == 0) ret += 8;
    EUDEndIf()
    if EUDIf()(_ATTC(b.p[0], 'y') == 0):
        ret.__iadd__(8)
        # (Line 68) if (b.p[1].y == 4) ret += 16;
    EUDEndIf()
    if EUDIf()(_ATTC(b.p[1], 'y') == 4):
        ret.__iadd__(16)
        # (Line 69) if (a.q == 3) ret += 32;
    EUDEndIf()
    if EUDIf()(_ATTC(a, 'q') == 3):
        ret.__iadd__(32)
        # (Line 70) if (b.q == 2) ret += 64;
    EUDEndIf()
    if EUDIf()(_ATTC(b, 'q') == 2):
        ret.__iadd__(64)
        # (Line 72) return ret;
    EUDEndIf()
    EUDReturn(ret)
    # (Line 73) }
    # (Line 75) var ListCount = 0;

ListCount = _TYGV([None], lambda: [0])
# (Line 76) object List {
# (Line 77) var prev: selftype, next: selftype;
class List(EUDStruct):
    # (Line 78) function constructor() {
    @EUDMethod
    def constructor(this):
        # (Line 79) ListCount += 1;
        ListCount.__iadd__(1)
        # (Line 80) }
        # (Line 81) function foo(c: Coord) {}

    @EUDTypedMethod([Coord])
    def foo(this, c):
        # (Line 82) };
        pass

    # (Line 83) function test_selftype_member() {
    _fields_ = [
        ('prev', selftype),
        ('next', selftype),
    ]

@EUDFunc
def f_test_selftype_member():
    # (Line 84) const a, b = List(), List();
    a, b = List2Assignable([List(), List()])
    # (Line 85) a.prev = b;
    _ATTW(a, 'prev') << (b)
    # (Line 86) a.next = b;
    _ATTW(a, 'next') << (b)
    # (Line 87) b.prev = a;
    _ATTW(b, 'prev') << (a)
    # (Line 88) b.next = a;
    _ATTW(b, 'next') << (a)
    # (Line 89) a.foo(0);
    a.foo(0)
    # (Line 90) const c = List();
    c = List()
    # (Line 91) const d = List();
    d = List()
    # (Line 92) const e = List();
    e = List()
    # (Line 94) return ListCount;
    EUDReturn(ListCount)
    # (Line 95) }
    # (Line 97) object RandomData {

# (Line 98) var arr: EUDArray;
class RandomData(EUDStruct):
    # (Line 99) var count;
    # (Line 101) function constructor(arr, count) {
    @EUDMethod
    def constructor(this, arr, count):
        # (Line 102) this.arr = arr;
        _ATTW(this, 'arr') << (arr)
        # (Line 103) this.count = count;
        _ATTW(this, 'count') << (count)
        # (Line 104) }
        # (Line 106) function get_value() {

    @EUDMethod
    def get_value(this):
        # (Line 107) return this.arr[dwrand() % this.count];
        EUDReturn(this.arr[f_dwrand() % this.count])
        # (Line 108) }
        # (Line 109) };

    # (Line 110) function test_eudmethods() {
    _fields_ = [
        ('arr', EUDArray),
        'count',
    ]

@EUDFunc
def f_test_eudmethods():
    # (Line 111) const rd = RandomData(EUDArray(list(1,2,3,4,5,6,7)), 7);
    rd = RandomData(EUDArray(FlattenList([1, 2, 3, 4, 5, 6, 7])), 7)
    # (Line 112) var ret = 0;
    ret = _TYLV([None], [0])
    # (Line 113) foreach(n : py_range(7)) {
    for n in range(7):
        # (Line 114) const v = rd.get_value();
        v = rd.get_value()
        # (Line 115) if (1 <= v && v <= 7) ret += 1;
        if EUDIf()(EUDSCAnd()(1 <= v)(v <= 7)()):
            ret.__iadd__(1)
            # (Line 116) }
        EUDEndIf()
        # (Line 117) return ret;

    EUDReturn(ret)
    # (Line 118) }



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import math

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...memio import f_dwread_epd
from ...memio.rwcommon import lv


@c.EUDFunc
def f_lengthdir(length, angle):
    # sin, cos table
    clist = []
    slist = []

    for i in range(91):
        cosv = math.floor(math.cos(math.pi / 180 * i) * 65536 + 0.5)
        sinv = math.floor(math.sin(math.pi / 180 * i) * 65536 + 0.5)
        clist.append(ut.i2b4(cosv))
        slist.append(ut.i2b4(sinv))

    cdb = c.Db(b"".join(clist))
    sdb = c.Db(b"".join(slist))

    # MAIN LOGIC

    if cs.EUDIf()(angle >= 360):
        angle %= 360
    cs.EUDEndIf()

    # sign of cos, sin
    sign = lv
    tableangle = c.EUDVariable()

    # get cos, sin from table
    if cs.EUDIf()(angle <= 89):
        c.VProc(angle, [sign.SetNumber(0), angle.QueueAssignTo(tableangle)])

    if cs.EUDElseIf()(angle <= 179):
        c.VProc(
            angle,
            [
                sign.SetNumber(1),
                tableangle.SetNumber(180),
                angle.QueueSubtractTo(tableangle),
            ],
        )

    if cs.EUDElseIf()(angle <= 269):
        c.VProc(
            angle,
            [
                sign.SetNumber(3),
                angle.QueueAddTo(tableangle),
                tableangle.SetNumber(-180),
            ],
        )

    if cs.EUDElse()():
        c.VProc(
            angle,
            [
                sign.SetNumber(2),
                angle.QueueSubtractTo(tableangle),
                tableangle.SetNumber(360),
            ],
        )

    cs.EUDEndIf()

    tablecos = f_dwread_epd(ut.EPD(cdb) + tableangle)
    tableangle += ut.EPD(sdb)
    tablesin = f_dwread_epd(tableangle)

    # calculate lengthdir: cos, sin * 65536
    ldir_x = tablecos
    ldir_x *= length
    ldir_x //= 65536

    ldir_y = tablesin
    ldir_y *= length
    ldir_y //= 65536

    # restore sign of cos, sin
    c.RawTrigger(
        conditions=sign.ExactlyX(1, 1),
        actions=[  # ldir_x = -ldir_x
            ldir_x.AddNumberX(0xFFFFFFFF, 0x55555555),
            ldir_x.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
            ldir_x.AddNumber(1),
        ],
    )

    c.RawTrigger(
        conditions=sign.ExactlyX(2, 2),
        actions=[  # ldir_y = -ldir_y
            ldir_y.AddNumberX(0xFFFFFFFF, 0x55555555),
            ldir_y.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
            ldir_y.AddNumber(1),
        ],
    )

    return ldir_x, ldir_y


@c.EUDFunc
def f_lengthdir_256(length, angle):
    # sin, cos table
    clist = []
    slist = []

    for i in range(65):
        cosv = math.floor(math.cos(math.pi / 128 * i) * 65536 + 0.5)
        sinv = math.floor(math.sin(math.pi / 128 * i) * 65536 + 0.5)
        clist.append(ut.i2b4(cosv))
        slist.append(ut.i2b4(sinv))

    cdb = c.Db(b"".join(clist))
    sdb = c.Db(b"".join(slist))

    # MAIN LOGIC
    cs.DoActions(angle.AddNumber(192), angle.SetNumberX(0, ~255))

    # sign of cos, sin
    sign = lv
    tableangle = c.EUDVariable()

    # get cos, sin from table
    if cs.EUDIf()(angle <= 63):
        c.VProc(angle, [sign.SetNumber(0), angle.QueueAssignTo(tableangle)])

    if cs.EUDElseIf()(angle <= 127):
        c.VProc(
            angle,
            [
                sign.SetNumber(1),
                tableangle.SetNumber(128),
                angle.QueueSubtractTo(tableangle),
            ],
        )

    if cs.EUDElseIf()(angle <= 191):
        c.VProc(
            angle,
            [
                sign.SetNumber(3),
                angle.QueueAddTo(tableangle),
                tableangle.SetNumber(-128),
            ],
        )

    if cs.EUDElse()():
        c.VProc(
            angle,
            [
                sign.SetNumber(2),
                angle.QueueSubtractTo(tableangle),
                tableangle.SetNumber(256),
            ],
        )

    cs.EUDEndIf()

    tablecos = f_dwread_epd(ut.EPD(cdb) + tableangle)
    tableangle += ut.EPD(sdb)
    tablesin = f_dwread_epd(tableangle)

    # calculate lengthdir: cos, sin * 65536
    ldir_x = tablecos
    ldir_x *= length
    ldir_x //= 65536

    ldir_y = tablesin
    ldir_y *= length
    ldir_y //= 65536

    # restore sign of cos, sin
    c.RawTrigger(
        conditions=sign.ExactlyX(1, 1),
        actions=[  # ldir_x = -ldir_x
            ldir_x.AddNumberX(0xFFFFFFFF, 0x55555555),
            ldir_x.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
            ldir_x.AddNumber(1),
        ],
    )

    c.RawTrigger(
        conditions=sign.ExactlyX(2, 2),
        actions=[  # ldir_y = -ldir_y
            ldir_y.AddNumberX(0xFFFFFFFF, 0x55555555),
            ldir_y.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
            ldir_y.AddNumber(1),
        ],
    )

    return ldir_x, ldir_y



# Copyright 2024 by zzt (Defender).
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar

from .. import utils as ut
from ..core.rawtrigger.consttype import ConstType
from ..core.rawtrigger.strenc import EncodeWeapon
from ..core.rawtrigger.typehint import _Weapon
from ..localize import _
from .offsetmap import (
    ByteMember,
    DamageTypeMember,
    DwordMember,
    EPDOffsetMap,
    ExplosionTypeMember,
    Flag,
    FlingyMember,
    IconMember,
    StatTextMember,
    UpgradeMember,
    WeaponBehaviorMember,
    WordEnumMember,
    WordMember,
)


class TargetFlags(WordEnumMember):
    __slots__ = ()
    Air = Flag(0x001)
    Ground = Flag(0x002)
    Mechanical = Flag(0x004)
    Organic = Flag(0x008)
    NonBuilding = Flag(0x010)
    NonRobotic = Flag(0x020)
    Terrain = Flag(0x040)
    OrganicOrMechanical = Flag(0x080)
    PlayerOwned = Flag(0x100)  #
    "Can target your own unit, used by Consume"


class Weapon(EPDOffsetMap, ConstType):
    __slots__ = ()
    label: ClassVar = StatTextMember("array", 0x6572E0)
    flingy: ClassVar = FlingyMember("array", 0x656CA8)
    # specialAttack: ClassVar = ByteMember("array", 0x6573E8)
    # special attack is for reference only (unused)
    targetFlags = TargetFlags("array", 0x657998)
    # can't use name 'range' because it's a python keyword
    minRange: ClassVar = DwordMember("array", 0x656A18)
    maxRange: ClassVar = DwordMember("array", 0x657470)
    upgrade: ClassVar = UpgradeMember("array", 0x6571D0)
    damageType: ClassVar[DamageTypeMember] = DamageTypeMember("array", 0x657258)
    behavior: ClassVar = WeaponBehaviorMember("array", 0x656670)
    "Fly and follow target, appear on target unit, etc."
    removeAfter: ClassVar = ByteMember("array", 0x657040)
    explosionType: ClassVar = ExplosionTypeMember("array", 0x6566F8)
    splashInnerRadius: ClassVar = WordMember("array", 0x656888)
    splashMiddleRadius: ClassVar = WordMember("array", 0x6570C8)
    splashOuterRadius: ClassVar = WordMember("array", 0x657780)
    damage: ClassVar = WordMember("array", 0x656EB0)
    damageBonus: ClassVar = WordMember("array", 0x657678)
    cooldown: ClassVar = ByteMember("array", 0x656FB8)
    damageFactor: ClassVar = ByteMember("array", 0x6564E0)
    "aka 'missile count'"
    attackAngle: ClassVar = ByteMember("array", 0x656990)
    launchSpin: ClassVar = ByteMember("array", 0x657888)
    forwardOffset: ClassVar = ByteMember("array", 0x657910)
    graphicXOffset: ClassVar = ByteMember("array", 0x657910)
    verticalOffset: ClassVar = ByteMember("array", 0x656C20)
    graphicYOffset: ClassVar = ByteMember("array", 0x656C20)
    targetErrorMessage: ClassVar = StatTextMember("array", 0x656568)
    icon: ClassVar = IconMember("array", 0x656780)

    @ut.classproperty
    def range(self):
        return (0, 129, 1)

    @classmethod
    def cast(cls, _from: _Weapon):
        if isinstance(_from, ConstType) and not isinstance(_from, cls):
            raise ut.EPError(_('"{}" is not a {}').format(_from, cls.__name__))
        return super().cast(_from)

    def __init__(self, initval: _Weapon) -> None:
        super().__init__(EncodeWeapon(initval))
